from typing import Final

from PyQt5.QtWidgets import (
    QFormLayout,
    QGridLayout,
    QGroupBox,
    QLabel,
    QSpinBox,
    QVBoxLayout,
    QWidget,
    QWizardPage,
)

from ptychodus.api.observer import Observable

from ....model.diffraction import DiffractionSettings, PatternSizer
from ....view.diffraction import OpenDatasetWizardPage

from ...data import FileDialogFactory
from ...parametric import (
    CheckBoxParameterViewController,
    CheckableGroupBoxParameterViewController,
    ParameterViewController,
    PathParameterViewController,
    SpinBoxParameterViewController,
)


class PatternMemoryMapViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self, settings: DiffractionSettings, file_dialog_factory: FileDialogFactory
    ) -> None:
        super().__init__(settings.memmap_enabled, 'Memory Map Diffraction Data')
        self._view_controller = PathParameterViewController.create_directory_chooser(
            settings.scratch_directory, file_dialog_factory
        )

        layout = QFormLayout()
        layout.addRow('Scratch Directory:', self._view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PatternCropViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        settings: DiffractionSettings,
        sizer: PatternSizer,
    ) -> None:
        super().__init__(settings.crop_enabled, 'Crop')
        self._settings = settings
        self._sizer = sizer

        self._center_x_spin_box = QSpinBox()
        self._center_y_spin_box = QSpinBox()
        self._width_spin_box = QSpinBox()
        self._height_spin_box = QSpinBox()

        layout = QGridLayout()
        layout.addWidget(QLabel('Center:'), 0, 0)
        layout.addWidget(self._center_x_spin_box, 0, 1)
        layout.addWidget(self._center_y_spin_box, 0, 2)
        layout.addWidget(QLabel('Extent:'), 1, 0)
        layout.addWidget(self._width_spin_box, 1, 1)
        layout.addWidget(self._height_spin_box, 1, 2)
        layout.setColumnStretch(1, 1)
        layout.setColumnStretch(2, 1)
        self.get_widget().setLayout(layout)

        self._sync_model_to_view()

        self._center_x_spin_box.valueChanged.connect(settings.crop_center_x_px.set_value)
        self._center_y_spin_box.valueChanged.connect(settings.crop_center_y_px.set_value)
        self._width_spin_box.valueChanged.connect(settings.crop_width_px.set_value)
        self._height_spin_box.valueChanged.connect(settings.crop_height_px.set_value)

        sizer.add_observer(self)

    def _sync_model_to_view(self) -> None:
        center_x = self._sizer.axis_x.get_crop_center()
        center_y = self._sizer.axis_y.get_crop_center()
        width = self._sizer.axis_x.get_crop_size()
        height = self._sizer.axis_y.get_crop_size()

        center_x_limits = self._sizer.axis_x.get_crop_center_limits()
        center_y_limits = self._sizer.axis_y.get_crop_center_limits()
        width_limits = self._sizer.axis_x.get_crop_size_limits()
        height_limits = self._sizer.axis_y.get_crop_size_limits()

        self._center_x_spin_box.blockSignals(True)
        self._center_x_spin_box.setRange(center_x_limits.lower, center_x_limits.upper)
        self._center_x_spin_box.setValue(center_x)
        self._center_x_spin_box.blockSignals(False)

        self._center_y_spin_box.blockSignals(True)
        self._center_y_spin_box.setRange(center_y_limits.lower, center_y_limits.upper)
        self._center_y_spin_box.setValue(center_y)
        self._center_y_spin_box.blockSignals(False)

        self._width_spin_box.blockSignals(True)
        self._width_spin_box.setRange(width_limits.lower, width_limits.upper)
        self._width_spin_box.setValue(width)
        self._width_spin_box.blockSignals(False)

        self._height_spin_box.blockSignals(True)
        self._height_spin_box.setRange(height_limits.lower, height_limits.upper)
        self._height_spin_box.setValue(height)
        self._height_spin_box.blockSignals(False)

    def _update(self, observable: Observable) -> None:
        if observable is self._sizer:
            self._sync_model_to_view()
        else:
            super()._update(observable)


class PatternBinningViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        settings: DiffractionSettings,
        sizer: PatternSizer,
    ) -> None:
        super().__init__(settings.binning_enabled, 'Bin Pixels')
        self._settings = settings
        self._sizer = sizer

        self._bin_size_x_spin_box = QSpinBox()
        self._bin_size_y_spin_box = QSpinBox()

        layout = QGridLayout()
        layout.addWidget(QLabel('Bin Size:'), 0, 0)
        layout.addWidget(self._bin_size_x_spin_box, 0, 1)
        layout.addWidget(self._bin_size_y_spin_box, 0, 2)
        layout.setColumnStretch(1, 1)
        layout.setColumnStretch(2, 1)
        self.get_widget().setLayout(layout)

        self._sync_model_to_view()

        self._bin_size_x_spin_box.valueChanged.connect(settings.bin_size_x.set_value)
        self._bin_size_y_spin_box.valueChanged.connect(settings.bin_size_y.set_value)

        sizer.add_observer(self)

    def _sync_model_to_view(self) -> None:
        bin_size_x = self._sizer.axis_x.get_bin_size()
        bin_size_y = self._sizer.axis_y.get_bin_size()

        bin_size_x_limits = self._sizer.axis_x.get_bin_size_limits()
        bin_size_y_limits = self._sizer.axis_y.get_bin_size_limits()

        self._bin_size_x_spin_box.blockSignals(True)
        self._bin_size_x_spin_box.setRange(bin_size_x_limits.lower, bin_size_x_limits.upper)
        self._bin_size_x_spin_box.setValue(bin_size_x)
        self._bin_size_x_spin_box.blockSignals(False)

        self._bin_size_y_spin_box.blockSignals(True)
        self._bin_size_y_spin_box.setRange(bin_size_y_limits.lower, bin_size_y_limits.upper)
        self._bin_size_y_spin_box.setValue(bin_size_y)
        self._bin_size_y_spin_box.blockSignals(False)

    def _update(self, observable: Observable) -> None:
        if observable is self._sizer:
            self._sync_model_to_view()
        else:
            super()._update(observable)


class PatternPaddingViewController(CheckableGroupBoxParameterViewController):
    MAX_INT: Final[int] = 0x7FFFFFFF

    def __init__(
        self,
        settings: DiffractionSettings,
        sizer: PatternSizer,
    ) -> None:
        super().__init__(settings.padding_enabled, 'Pad')
        self._settings = settings
        self._sizer = sizer

        self._pad_x_spin_box = QSpinBox()
        self._pad_y_spin_box = QSpinBox()

        layout = QGridLayout()
        layout.addWidget(QLabel('Padding:'), 0, 0)
        layout.addWidget(self._pad_x_spin_box, 0, 1)
        layout.addWidget(self._pad_y_spin_box, 0, 2)
        layout.setColumnStretch(1, 1)
        layout.setColumnStretch(2, 1)
        self.get_widget().setLayout(layout)

        self._sync_model_to_view()

        self._pad_x_spin_box.valueChanged.connect(settings.pad_x.set_value)
        self._pad_y_spin_box.valueChanged.connect(settings.pad_y.set_value)

        sizer.add_observer(self)

    def _sync_model_to_view(self) -> None:
        pad_x = self._sizer.axis_x.get_pad_size()
        pad_y = self._sizer.axis_y.get_pad_size()

        self._pad_x_spin_box.blockSignals(True)
        self._pad_x_spin_box.setRange(0, self.MAX_INT)
        self._pad_x_spin_box.setValue(pad_x)
        self._pad_x_spin_box.blockSignals(False)

        self._pad_y_spin_box.blockSignals(True)
        self._pad_y_spin_box.setRange(0, self.MAX_INT)
        self._pad_y_spin_box.setValue(pad_y)
        self._pad_y_spin_box.blockSignals(False)

    def _update(self, observable: Observable) -> None:
        if observable is self._sizer:
            self._sync_model_to_view()
        else:
            super()._update(observable)


class PatternTransformViewController:
    def __init__(
        self, settings: DiffractionSettings, file_dialog_factory: FileDialogFactory
    ) -> None:
        self._hflip_view_controller = CheckBoxParameterViewController(
            settings.hflip, 'Flip Horizontal'
        )
        self._vflip_view_controller = CheckBoxParameterViewController(
            settings.vflip, 'Flip Vertical'
        )
        self._transpose_view_controller = CheckBoxParameterViewController(
            settings.transpose, 'Transpose'
        )
        self._lower_bound_enabled_view_controller = CheckBoxParameterViewController(
            settings.value_lower_bound_enabled, 'Value Lower Bound:'
        )
        self._lower_bound_view_controller = SpinBoxParameterViewController(
            settings.value_lower_bound
        )
        self._upper_bound_enabled_view_controller = CheckBoxParameterViewController(
            settings.value_upper_bound_enabled, 'Value upper Bound:'
        )
        self._upper_bound_view_controller = SpinBoxParameterViewController(
            settings.value_upper_bound
        )

        layout = QGridLayout()
        layout.addWidget(QLabel('Axes:'), 0, 0)
        layout.addWidget(self._hflip_view_controller.get_widget(), 0, 1)
        layout.addWidget(self._vflip_view_controller.get_widget(), 0, 2)
        layout.addWidget(self._transpose_view_controller.get_widget(), 0, 3)
        layout.addWidget(self._lower_bound_enabled_view_controller.get_widget(), 1, 0)
        layout.addWidget(self._lower_bound_view_controller.get_widget(), 1, 1, 1, 3)
        layout.addWidget(self._upper_bound_enabled_view_controller.get_widget(), 2, 0)
        layout.addWidget(self._upper_bound_view_controller.get_widget(), 2, 1, 1, 3)
        layout.setColumnStretch(1, 1)
        layout.setColumnStretch(2, 1)
        layout.setColumnStretch(3, 1)

        self._widget = QGroupBox('Transform')
        self._widget.setLayout(layout)

    def get_widget(self) -> QWidget:
        return self._widget


class OpenDatasetWizardPatternsViewController(ParameterViewController):
    def __init__(
        self,
        settings: DiffractionSettings,
        sizer: PatternSizer,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        self._memory_map_view_controller = PatternMemoryMapViewController(
            settings, file_dialog_factory
        )
        self._crop_view_controller = PatternCropViewController(settings, sizer)
        self._binning_view_controller = PatternBinningViewController(settings, sizer)
        self._padding_view_controller = PatternPaddingViewController(settings, sizer)
        self._transform_view_controller = PatternTransformViewController(
            settings, file_dialog_factory
        )

        layout = QVBoxLayout()
        layout.addWidget(self._memory_map_view_controller.get_widget())
        layout.addWidget(self._crop_view_controller.get_widget())
        layout.addWidget(self._binning_view_controller.get_widget())
        layout.addWidget(self._padding_view_controller.get_widget())
        layout.addWidget(self._transform_view_controller.get_widget())
        layout.addStretch()

        self._page = OpenDatasetWizardPage()
        self._page.setTitle('Pattern Processing')
        self._page._set_complete(True)
        self._page.setLayout(layout)

    def get_widget(self) -> QWizardPage:
        return self._page



from __future__ import annotations
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any
import logging

import numpy

from ptychodus.api.common import RealArrayType
from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.object import ObjectCenter
from ptychodus.api.observer import Observable

from ..product import ProductRepository
from .interpolators import BarycentricArrayStitcher


__all__ = [
    'IlluminationMapper',
]

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class IlluminationMap:
    photon_number: RealArrayType
    photon_flux_Hz: float  # noqa: N815
    photon_energy_J: float  # noqa: N815
    exposure_time_s: float
    mass_attenuation_m2_kg: float
    pixel_geometry: PixelGeometry
    center: ObjectCenter

    @property
    def photon_fluence_1_m2(self) -> RealArrayType:
        return self.photon_number / self.pixel_geometry.get_area_m2()

    @property
    def photon_fluence_rate_Hz_m2(self) -> RealArrayType:  # noqa: N802
        return self.photon_fluence_1_m2 / self.exposure_time_s

    @property
    def energy_fluence_J_m2(self) -> RealArrayType:  # noqa: N802
        return self.photon_fluence_1_m2 * self.photon_energy_J

    @property
    def energy_fluence_rate_W_m2(self) -> RealArrayType:  # noqa: N802
        return self.photon_fluence_rate_Hz_m2 * self.photon_energy_J

    @property
    def dose_Gy(self) -> RealArrayType:  # noqa: N802
        return self.energy_fluence_J_m2 * self.mass_attenuation_m2_kg

    @property
    def dose_rate_Gy_s(self) -> RealArrayType:  # noqa: N802
        return self.energy_fluence_rate_W_m2 * self.mass_attenuation_m2_kg

    @property
    def intensity_W_m2(self) -> RealArrayType:  # noqa: N802
        return self.energy_fluence_rate_W_m2


class IlluminationMapper(Observable):
    def __init__(self, repository: ProductRepository) -> None:
        super().__init__()
        self._repository = repository

        self._product_index = -1
        self._illumination_map: IlluminationMap | None = None

    def set_product(self, product_index: int) -> None:
        if self._product_index != product_index:
            self._product_index = product_index
            self._illumination_map = None
            self.notify_observers()

    def get_product_name(self) -> str:
        product = self._repository[self._product_index]
        return product.get_name()

    def map(self) -> None:
        product = self._repository[self._product_index].get_product()
        object_geometry = product.object_.get_geometry()

        stitcher = BarycentricArrayStitcher[numpy.floating[Any]](
            numpy.zeros((object_geometry.height_px, object_geometry.width_px))
        )

        for scan_point, probe in zip(product.probe_positions, product.probes):
            object_point = object_geometry.map_coordinates_probe_to_object(scan_point)
            stitcher.add_patch(
                object_point.coordinate_x_px,
                object_point.coordinate_y_px,
                probe.get_intensity(),
            )

        exposure_time_s = product.metadata.exposure_time_s
        photon_flux_Hz = float('nan')  # noqa: N806

        try:
            photon_flux_Hz = product.metadata.probe_photon_count / exposure_time_s  # noqa: N806
        except ZeroDivisionError:
            pass

        self._illumination_map = IlluminationMap(
            photon_number=stitcher.stitch(),
            photon_flux_Hz=photon_flux_Hz,
            photon_energy_J=product.metadata.probe_energy_J,
            exposure_time_s=exposure_time_s,
            mass_attenuation_m2_kg=product.metadata.mass_attenuation_m2_kg,
            pixel_geometry=object_geometry.get_pixel_geometry(),
            center=object_geometry.get_center(),
        )
        self.notify_observers()

    def get_illumination_map(self) -> IlluminationMap:
        if self._illumination_map is None:
            raise ValueError('No analyzed data!')

        return self._illumination_map

    def get_save_file_filters(self) -> Sequence[str]:
        return [self.get_save_file_filter()]

    def get_save_file_filter(self) -> str:
        return 'NumPy Zipped Archive (*.npz)'

    def save_data(self, file_path: Path) -> None:
        if self._illumination_map is None:
            raise ValueError('No analyzed data!')

        contents: dict[str, Any] = {
            'photon_number': self._illumination_map.photon_number,
            'photon_fluence_1_m2': self._illumination_map.photon_fluence_1_m2,
            'photon_fluence_rate_Hz_m2': self._illumination_map.photon_fluence_rate_Hz_m2,
            'energy_fluence_J_m2': self._illumination_map.energy_fluence_J_m2,
            'energy_fluence_rate_W_m2': self._illumination_map.energy_fluence_rate_W_m2,
            'dose_Gy': self._illumination_map.dose_Gy,
            'dose_rate_Gy_s': self._illumination_map.dose_rate_Gy_s,
            'pixel_height_m': self._illumination_map.pixel_geometry.height_m,
            'pixel_width_m': self._illumination_map.pixel_geometry.width_m,
            'center_x_m': self._illumination_map.center.coordinate_x_m,
            'center_y_m': self._illumination_map.center.coordinate_y_m,
        }

        numpy.savez_compressed(file_path, allow_pickle=False, **contents)



from __future__ import annotations
from collections.abc import Mapping, Sequence
from pprint import pformat
from typing import Final, TypeAlias
import logging
import os

import fair_research_login
import gladier
import gladier.managers
import globus_sdk

AuthorizerTypes: TypeAlias = globus_sdk.AccessTokenAuthorizer | globus_sdk.RefreshTokenAuthorizer
ScopeAuthorizerMapping: TypeAlias = Mapping[str, AuthorizerTypes]

PTYCHODUS_CLIENT_ID: Final[str] = '5c0fb474-ae53-44c2-8c32-dd0db9965c57'

logger = logging.getLogger(__name__)


def process_with_ptychodus(**data: str) -> None:
    from pathlib import Path
    from ptychodus.model import ModelCore

    action = data['ptychodus_action']
    input_directory = Path(data['ptychodus_input_directory'])
    output_directory = Path(data['ptychodus_output_directory'])

    with ModelCore() as model:
        model.batch_mode_execute(action, input_directory, output_directory)


@gladier.generate_flow_definition
class PtychodusProcessingTool(gladier.GladierBaseTool):
    compute_functions = [process_with_ptychodus]
    required_input = [
        'ptychodus_action',
        'ptychodus_input_directory',
        'ptychodus_output_directory',
    ]


@gladier.generate_flow_definition
class PtychodusClient(gladier.GladierBaseClient):
    client_id = PTYCHODUS_CLIENT_ID
    gladier_tools = [
        'gladier_tools.globus.transfer.Transfer:InputData',
        PtychodusProcessingTool,
        'gladier_tools.globus.transfer.Transfer:OutputData',
        # TODO 'gladier_tools.publish.Publish',
    ]

    @classmethod
    def _create_confidential_client(cls, client_id: str, client_secret: str) -> PtychodusClient:
        auth_client = globus_sdk.ConfidentialAppAuthClient(
            client_id=client_id,
            client_secret=client_secret,
            app_name='Ptychodus',
        )

        def _request_auth(scopes: list[str]) -> ScopeAuthorizerMapping:
            logger.debug(f'Requested authorization scopes: {pformat(scopes)}')

            response = auth_client.oauth2_client_credentials_tokens(requested_scopes=scopes)
            return {
                scope: globus_sdk.AccessTokenAuthorizer(access_token=tokens['access_token'])
                for scope, tokens in response.by_scopes.scope_map.items()
            }

        initial_authorizers: dict[str, AuthorizerTypes] = dict()
        login_manager = gladier.managers.CallbackLoginManager(
            authorizers=initial_authorizers,
            callback=_request_auth,
        )

        flow_id = os.environ.get('FLOW_ID', '')

        if not flow_id:
            # This isn't necessarily bad, but confidential clients (like regular
            # users) only get one flow to play with. They probably don't need
            # more than one, but this will ensure there aren't errors due to
            # tracking mismatch in the Glaider config.
            logger.warning('No flow ID enforced. Recommend setting FLOW_ID environment variable.')

        flows_manager = gladier.managers.FlowsManager(flow_id=flow_id)
        return cls(login_manager=login_manager, flows_manager=flows_manager)

    @classmethod
    def _create_native_client(
        cls, code_handlers: Sequence[fair_research_login.CodeHandler]
    ) -> PtychodusClient:
        auth_client = fair_research_login.NativeClient(
            client_id=PTYCHODUS_CLIENT_ID,
            app_name='Ptychodus',
            code_handlers=code_handlers,
        )

        def _request_auth(self, requested_scopes: list[str]) -> ScopeAuthorizerMapping:
            logger.debug(f'Requested authorization scopes: {pformat(requested_scopes)}')

            # 'force' is used for any underlying scope changes. For example, if a flow adds transfer
            # functionality since it was last run, running it again would require a re-login.
            auth_client.login(requested_scopes=requested_scopes, force=True, refresh_tokens=True)
            return auth_client.get_authorizers_by_scope()  # type: ignore

        initial_authorizers: dict[str, AuthorizerTypes] = dict()

        try:
            # Try to use a previous login to avoid a new login flow
            initial_authorizers = auth_client.get_authorizers_by_scope()  # type: ignore
        except fair_research_login.LoadError:
            pass

        login_manager = gladier.managers.CallbackLoginManager(
            authorizers=initial_authorizers,
            callback=_request_auth,
        )
        return cls(login_manager=login_manager)

    @classmethod
    def create_client(
        cls, code_handlers: Sequence[fair_research_login.CodeHandler]
    ) -> PtychodusClient:
        try:
            client_id = os.environ['CLIENT_ID']
            client_secret = os.environ['CLIENT_SECRET']
        except KeyError:
            logger.info(
                'Environment variables "CLIENT_ID" and "CLIENT_SECRET" not set. '
                'Creating native PtychodusClient.'
            )
            return cls._create_native_client(code_handlers)
        else:
            logger.info(
                'Environment variables "CLIENT_ID" and "CLIENT_SECRET" set. '
                'Creating confidential PtychodusClient.'
            )
            return cls._create_confidential_client(client_id, client_secret)



"""Object (transmission function) data structures and file I/O plugin interfaces."""

from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path

import numpy

from .common import ComplexArrayType
from .geometry import PixelGeometry
from .probe_positions import ProbePosition


@dataclass(frozen=True)
class ObjectCenter:
    """Physical center coordinates of the object array in meters."""

    coordinate_x_m: float
    coordinate_y_m: float

    def copy(self) -> ObjectCenter:
        return ObjectCenter(
            coordinate_x_m=float(self.coordinate_x_m),
            coordinate_y_m=float(self.coordinate_y_m),
        )


@dataclass(frozen=True)
class ObjectPosition:
    """Position expressed in object pixel coordinates."""

    index: int
    coordinate_x_px: float
    coordinate_y_px: float


@dataclass(frozen=True)
class ObjectGeometry:
    """Spatial geometry of the object: size, pixel scale, and center."""

    width_px: int
    height_px: int
    pixel_width_m: float
    pixel_height_m: float
    center_x_m: float
    center_y_m: float

    @property
    def width_m(self) -> float:
        return self.width_px * self.pixel_width_m

    @property
    def height_m(self) -> float:
        return self.height_px * self.pixel_height_m

    @property
    def minimum_x_m(self) -> float:
        return self.center_x_m - self.width_m / 2.0

    @property
    def minimum_y_m(self) -> float:
        return self.center_y_m - self.height_m / 2.0

    def get_pixel_geometry(self) -> PixelGeometry:
        return PixelGeometry(
            width_m=self.pixel_width_m,
            height_m=self.pixel_height_m,
        )

    def get_center(self) -> ObjectCenter:
        return ObjectCenter(
            coordinate_x_m=self.center_x_m,
            coordinate_y_m=self.center_y_m,
        )

    def map_coordinates_object_to_probe(self, position: ObjectPosition) -> ProbePosition:
        rx_px = self.width_px / 2
        ry_px = self.height_px / 2
        dx_m = self.pixel_width_m
        dy_m = self.pixel_height_m

        x_m = self.center_x_m + dx_m * (position.coordinate_x_px - rx_px)
        y_m = self.center_y_m + dy_m * (position.coordinate_y_px - ry_px)

        return ProbePosition(position.index, x_m, y_m)

    def map_coordinates_probe_to_object(self, position: ProbePosition) -> ObjectPosition:
        rx_px = self.width_px / 2
        ry_px = self.height_px / 2
        dx_m = self.pixel_width_m
        dy_m = self.pixel_height_m

        x_px = (position.coordinate_x_m - self.center_x_m) / dx_m + rx_px
        y_px = (position.coordinate_y_m - self.center_y_m) / dy_m + ry_px

        return ObjectPosition(position.index, x_px, y_px)

    def contains(self, geometry: ObjectGeometry) -> bool:
        dx = self.center_x_m - geometry.center_x_m
        dy = self.center_y_m - geometry.center_y_m
        dw = self.width_m - geometry.width_m
        dh = self.height_m - geometry.height_m
        return abs(dx) <= dw and abs(dy) <= dh


class ObjectGeometryProvider(ABC):
    """Interface for classes that provide object geometry."""

    @abstractmethod
    def get_probe_positions(self) -> Sequence[ProbePosition]:
        pass

    @abstractmethod
    def get_object_geometry(self) -> ObjectGeometry:
        pass


class Object:
    """Complex transmission function stored as a (layers, height, width) array with spatial metadata."""

    def __init__(
        self,
        array: ComplexArrayType | None,
        pixel_geometry: PixelGeometry | None,
        center: ObjectCenter | None,
        layer_spacing_m: Sequence[float] = [],
    ) -> None:
        if array is None:
            self._array: ComplexArrayType = numpy.zeros((1, 0, 0), dtype=complex)
        elif numpy.iscomplexobj(array):
            match array.ndim:
                case 2:
                    self._array = array[numpy.newaxis, ...]
                case 3:
                    self._array = array
                case _:
                    raise ValueError('Object must be 2- or 3-dimensional ndarray.')
        else:
            raise TypeError('Object must be a complex-valued ndarray')

        self._pixel_geometry = pixel_geometry
        self._center = center
        self._layer_spacing_m = layer_spacing_m

        expected_layers = self._array.shape[-3]
        actual_layers = len(layer_spacing_m) + 1

        if actual_layers != expected_layers:
            raise ValueError(f'Expected {expected_layers} layers; got {actual_layers}!')

    def copy(self) -> Object:
        return Object(
            array=self._array.copy(),
            pixel_geometry=None if self._pixel_geometry is None else self._pixel_geometry.copy(),
            center=None if self._center is None else self._center.copy(),
            layer_spacing_m=list(self._layer_spacing_m),
        )

    def get_array(self) -> ComplexArrayType:
        return self._array

    @property
    def dtype(self) -> numpy.dtype:
        return self._array.dtype

    @property
    def nbytes(self) -> int:
        return self._array.nbytes

    @property
    def width_px(self) -> int:
        return self._array.shape[-1]

    @property
    def height_px(self) -> int:
        return self._array.shape[-2]

    @property
    def num_layers(self) -> int:
        return self._array.shape[-3]

    def get_pixel_geometry(self) -> PixelGeometry:
        if self._pixel_geometry is None:
            raise ValueError('Missing object pixel geometry!')

        return self._pixel_geometry

    def get_center(self) -> ObjectCenter:
        if self._center is None:
            raise ValueError('Missing object center!')

        return self._center

    def get_geometry(self) -> ObjectGeometry:
        pixel_geometry = self.get_pixel_geometry()
        center = self.get_center()

        return ObjectGeometry(
            width_px=self.width_px,
            height_px=self.height_px,
            pixel_width_m=pixel_geometry.width_m,
            pixel_height_m=pixel_geometry.height_m,
            center_x_m=center.coordinate_x_m,
            center_y_m=center.coordinate_y_m,
        )

    def get_layer(self, number: int) -> ComplexArrayType:
        return self._array[number, :, :]

    def get_layers_flattened(self) -> ComplexArrayType:
        return numpy.prod(self._array, axis=-3)

    @property
    def layer_spacing_m(self) -> Sequence[float]:
        return self._layer_spacing_m

    def get_total_thickness_m(self) -> float:
        return sum(self._layer_spacing_m)

    def __repr__(self) -> str:
        return f'{self._array.dtype}{self._array.shape}'


class ObjectFileReader(ABC):
    """Plugin interface for reading objects."""

    @abstractmethod
    def read(self, file_path: Path) -> Object:
        """Read an object from file."""
        pass


class ObjectFileWriter(ABC):
    """Plugin interface for writing objects."""

    @abstractmethod
    def write(self, file_path: Path, object_: Object) -> None:
        """Write an object to file."""
        pass



from pathlib import Path

import numpy

from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe import ProbeSequence, ProbeFileReader, ProbeFileWriter


class CSVProbeFileReader(ProbeFileReader):
    def read(self, file_path: Path) -> ProbeSequence:
        array_flat = numpy.genfromtxt(file_path, delimiter=',', dtype=complex)
        num_modes, remainder = divmod(array_flat.shape[0], array_flat.shape[1])

        if remainder != 0:
            raise ValueError('Failed to determine probe modes!')

        array = array_flat.reshape(num_modes, array_flat.shape[1], array_flat.shape[1])
        return ProbeSequence(array=array, opr_weights=None, pixel_geometry=None)


class CSVProbeFileWriter(ProbeFileWriter):
    def write(self, file_path: Path, probes: ProbeSequence) -> None:
        array = probes.get_probe_no_opr().get_array()
        array_flat = array.reshape(-1, array.shape[-1])
        numpy.savetxt(file_path, array_flat, delimiter=',')


def register_plugins(registry: PluginRegistry) -> None:
    registry.probe_file_readers.register_plugin(
        CSVProbeFileReader(),
        simple_name='CSV',
        display_name='Comma-Separated Values Files (*.csv)',
    )
    registry.probe_file_writers.register_plugin(
        CSVProbeFileWriter(),
        simple_name='CSV',
        display_name='Comma-Separated Values Files (*.csv)',
    )



from PyQt5.QtWidgets import QFormLayout

from ptychodus.api.parametric import (
    BooleanParameter,
    IntegerParameter,
    RealParameter,
    StringParameter,
)

from ...model.ptychi import (
    PtyChiDMSettings,
    PtyChiEnumerators,
    PtyChiLSQMLSettings,
    PtyChiPIESettings,
    PtyChiProbeSettings,
)
from ..parametric import (
    CheckableGroupBoxParameterViewController,
    ComboBoxParameterViewController,
    DecimalLineEditParameterViewController,
    DecimalSliderParameterViewController,
)
from .optimizer import PtyChiOptimizationPlanViewController, PtyChiOptimizerParameterViewController

__all__ = ['PtyChiProbeViewController']


class PtyChiConstrainProbePowerViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        constrain_power: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        num_epochs: IntegerParameter,
    ) -> None:
        super().__init__(
            constrain_power, 'Constrain Power', tool_tip='Whether to constrain probe power'
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiOrthogonalizeIncoherentModesViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        orthogonalize_modes: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        method: StringParameter,
        num_epochs: IntegerParameter,
        enumerators: PtyChiEnumerators,
    ) -> None:
        super().__init__(
            orthogonalize_modes,
            'Orthogonalize Incoherent Modes',
            tool_tip='Whether to orthogonalize incoherent probe modes',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )
        self._method_view_controller = ComboBoxParameterViewController(
            method,
            enumerators.orthogonalization_methods(),
            tool_tip='Method to use for incoherent mode orthogonalization',
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        layout.addRow('Method:', self._method_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiOrthogonalizeOPRModesViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        orthogonalize_modes: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        num_epochs: IntegerParameter,
    ) -> None:
        super().__init__(
            orthogonalize_modes,
            'Orthogonalize OPR Modes',
            tool_tip='Whether to orthogonalize OPR modes',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiConstrainSupportViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        constrain_support: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        threshold: RealParameter,
        num_epochs: IntegerParameter,
    ) -> None:
        super().__init__(
            constrain_support,
            'Constrain Support',
            tool_tip='When enabled, the probe will be shrinkwrapped so that small values are set to zero',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )
        self._threshold_view_controller = DecimalLineEditParameterViewController(
            threshold, tool_tip='Threshold for the probe support constraint'
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        layout.addRow('Threshold:', self._threshold_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiConstrainCenterViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        constrain_center: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        num_epochs: IntegerParameter,
        use_intensity_for_mass_centroid: BooleanParameter,
    ) -> None:
        super().__init__(
            constrain_center,
            'Constrain Center',
            tool_tip='When enabled, the probe center of mass will be constrained to the center of the probe array',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )
        self._use_intensity_for_mass_centroid_view_controller = CheckableGroupBoxParameterViewController(
            use_intensity_for_mass_centroid,
            'Use Intensity for Mass Centroid',
            tool_tip='When enabled, the mass centroid will be calculated using the intensity of the probe',
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        layout.addRow(self._use_intensity_for_mass_centroid_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiProbeViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        settings: PtyChiProbeSettings,
        dm_settings: PtyChiDMSettings | None,
        lsqml_settings: PtyChiLSQMLSettings | None,
        pie_settings: PtyChiPIESettings | None,
        num_epochs: IntegerParameter,
        enumerators: PtyChiEnumerators,
    ) -> None:
        super().__init__(
            settings.is_optimizable, 'Optimize Probe', tool_tip='Whether the probe is optimizable'
        )
        self._optimization_plan_view_controller = PtyChiOptimizationPlanViewController(
            settings.optimization_plan_start,
            settings.optimization_plan_stop,
            settings.optimization_plan_stride,
            num_epochs,
        )
        self._optimizer_view_controller = PtyChiOptimizerParameterViewController(
            settings.optimizer, enumerators
        )
        self._step_size_view_controller = DecimalLineEditParameterViewController(
            settings.step_size, tool_tip='Optimizer step size'
        )
        self._constrain_probe_power_view_controller = PtyChiConstrainProbePowerViewController(
            settings.constrain_probe_power,
            settings.constrain_probe_power_start,
            settings.constrain_probe_power_stop,
            settings.constrain_probe_power_stride,
            num_epochs,
        )
        self._orthogonalize_incoherent_modes_view_controller = (
            PtyChiOrthogonalizeIncoherentModesViewController(
                settings.orthogonalize_incoherent_modes,
                settings.orthogonalize_incoherent_modes_start,
                settings.orthogonalize_incoherent_modes_stop,
                settings.orthogonalize_incoherent_modes_stride,
                settings.orthogonalize_incoherent_modes_method,
                num_epochs,
                enumerators,
            )
        )
        self._orthogonalize_opr_modes_view_controller = PtyChiOrthogonalizeOPRModesViewController(
            settings.orthogonalize_opr_modes,
            settings.orthogonalize_opr_modes_start,
            settings.orthogonalize_opr_modes_stop,
            settings.orthogonalize_opr_modes_stride,
            num_epochs,
        )
        self._constrain_support_view_controller = PtyChiConstrainSupportViewController(
            settings.constrain_support,
            settings.constrain_support_start,
            settings.constrain_support_stop,
            settings.constrain_support_stride,
            settings.constrain_support_threshold,
            num_epochs,
        )
        self._constrain_center_view_controller = PtyChiConstrainCenterViewController(
            settings.constrain_center,
            settings.constrain_center_start,
            settings.constrain_center_stop,
            settings.constrain_center_stride,
            num_epochs,
            settings.use_intensity_for_mass_centroid,
        )
        self._relax_eigenmode_update_view_controller = DecimalSliderParameterViewController(
            settings.relax_eigenmode_update,
            tool_tip='Relaxation factor for the eigenmode update',
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._optimization_plan_view_controller.get_widget())
        layout.addRow('Optimizer:', self._optimizer_view_controller.get_widget())
        layout.addRow('Step Size:', self._step_size_view_controller.get_widget())
        layout.addRow(self._constrain_probe_power_view_controller.get_widget())
        layout.addRow(self._orthogonalize_incoherent_modes_view_controller.get_widget())
        layout.addRow(self._orthogonalize_opr_modes_view_controller.get_widget())
        layout.addRow(self._constrain_support_view_controller.get_widget())
        layout.addRow(self._constrain_center_view_controller.get_widget())
        layout.addRow(
            'Relax Eigenmode Update:', self._relax_eigenmode_update_view_controller.get_widget()
        )

        if dm_settings is not None:
            self._inertia_view_controller = DecimalLineEditParameterViewController(
                dm_settings.probe_inertia, tool_tip='Inertia for the probe update'
            )
            layout.addRow('Inertia:', self._inertia_view_controller.get_widget())

        if lsqml_settings is not None:
            self._optimal_step_size_scaler_view_controller = DecimalLineEditParameterViewController(
                lsqml_settings.probe_optimal_step_size_scaler,
                tool_tip='Optimal step size scaler for the probe update',
            )
            layout.addRow(
                'Optimal Step Size Scaler:',
                self._optimal_step_size_scaler_view_controller.get_widget(),
            )

        if pie_settings is not None:
            self._alpha = DecimalSliderParameterViewController(
                pie_settings.probe_alpha, tool_tip='Relaxation factor for the probe update'
            )
            layout.addRow('Alpha:', self._alpha.get_widget())

        self.get_widget().setLayout(layout)



import logging

import numpy

from ptychodus.api.object import Object, ObjectGeometryProvider

from .builder import FromMemoryObjectBuilder
from .builder_factory import ObjectBuilderFactory
from .item import ObjectRepositoryItem
from .settings import ObjectSettings

logger = logging.getLogger(__name__)


class ObjectRepositoryItemFactory:
    def __init__(
        self,
        rng: numpy.random.Generator,
        settings: ObjectSettings,
        builder_factory: ObjectBuilderFactory,
    ) -> None:
        self._rng = rng
        self._settings = settings
        self._builder_factory = builder_factory

    def create(
        self, geometry_provider: ObjectGeometryProvider, object_: Object | None = None
    ) -> ObjectRepositoryItem:
        # TODO layers_builder = MultilayerObjectBuilder()

        if object_ is None:
            builder = self._builder_factory.create_default()
        else:
            builder = FromMemoryObjectBuilder(self._settings, object_)
            # TODO layers_builder.set_identity()

        return ObjectRepositoryItem(geometry_provider, self._settings, builder)

    def create_from_settings(
        self, geometry_provider: ObjectGeometryProvider
    ) -> ObjectRepositoryItem:
        try:
            builder = self._builder_factory.create_from_settings()
        except Exception as exc:
            logger.error(''.join(exc.args))
            builder = self._builder_factory.create_default()

        return ObjectRepositoryItem(geometry_provider, self._settings, builder)



from collections.abc import Iterator, Sequence
from pathlib import Path
from typing import Any, Final
import logging

import numpy

from ptycho.config.config import InferenceConfig, ModelConfig, TrainingConfig, update_legacy_dict
from ptycho.raw_data import RawData
from ptycho.workflows.components import load_inference_bundle
import ptycho.loader
import ptycho.model_manager
import ptycho.params

from ptychodus.api.object import Object
from ptychodus.api.product import Product
from ptychodus.api.reconstructor import (
    LossValue,
    ReconstructInput,
    ReconstructOutput,
    TrainOutput,
    TrainableReconstructor,
)

from .settings import (
    PtychoPINNInferenceSettings,
    PtychoPINNModelSettings,
    PtychoPINNTrainingSettings,
)

__all__ = [
    'PtychoPINNTrainableReconstructor',
]

logger = logging.getLogger(__name__)


def create_raw_data(parameters: ReconstructInput) -> RawData:
    object_geometry = parameters.product.object_.get_geometry()
    position_x_px: list[float] = list()
    position_y_px: list[float] = list()

    for scan_point in parameters.product.probe_positions:
        object_point = object_geometry.map_coordinates_probe_to_object(scan_point)
        position_x_px.append(object_point.coordinate_x_px)
        position_y_px.append(object_point.coordinate_y_px)

    return RawData.from_coords_without_pc(
        xcoords=numpy.array(position_x_px),
        ycoords=numpy.array(position_y_px),
        diff3d=parameters.diffraction_patterns,
        probeGuess=parameters.product.probes.get_probe_no_opr().get_incoherent_mode(0),
        # assume that all patches are from the same object
        scan_index=numpy.zeros(len(parameters.product.probe_positions), dtype=int),
        objectGuess=parameters.product.object_.get_layer(0),
    )


class PtychoPINNTrainableReconstructor(TrainableReconstructor):
    MODEL_FILE_NAME: Final[str] = 'wts.h5.zip'

    def __init__(
        self,
        name: str,
        model_settings: PtychoPINNModelSettings,
        inference_settings: PtychoPINNInferenceSettings,
        training_settings: PtychoPINNTrainingSettings,
        *,
        is_developer_mode_enabled: bool,
    ) -> None:
        super().__init__()
        self._name = name
        self._model_settings = model_settings
        self._inference_settings = inference_settings
        self._training_settings = training_settings
        self.__model: Any = None
        self._config: dict[str, Any] = dict()
        self._is_developer_mode_enabled = is_developer_mode_enabled

    def _create_model_config(self, model_size: int) -> ModelConfig:
        return ModelConfig(
            N=model_size,
            gridsize=self._model_settings.gridsize.get_value(),
            n_filters_scale=self._model_settings.n_filters_scale.get_value(),
            model_type=self._name.lower(),
            amp_activation=self._model_settings.amp_activation.get_value(),
            object_big=self._model_settings.object_big.get_value(),
            probe_big=self._model_settings.probe_big.get_value(),
            probe_mask=self._model_settings.probe_mask.get_value(),
            pad_object=self._model_settings.pad_object.get_value(),
            probe_scale=self._model_settings.probe_scale.get_value(),
            gaussian_smoothing_sigma=self._model_settings.gaussian_smoothing_sigma.get_value(),
        )

    @property
    def name(self) -> str:
        return self._name

    def get_progress_goal(self) -> int:
        return 0

    @property
    def _model(self) -> Any:  # TODO tensorflow.keras.Model | None
        if self.__model is None:
            raise RuntimeError('Model not loaded!')

        return self.__model

    def _reconstruct_image(self, test_data: ptycho.loader.PtychoDataContainer) -> Any:
        try:
            intensity_scale = ptycho.params.get('intensity_scale')
        except KeyError as exc:
            raise RuntimeError('Missing intensity_scale in ptycho.params.cfg') from exc
        return self._model.predict([test_data.X * intensity_scale, test_data.local_offsets])

    def reconstruct(self, parameters: ReconstructInput) -> Iterator[ReconstructOutput]:
        model_size = parameters.diffraction_patterns.shape[-1]

        if parameters.diffraction_patterns.shape[-2] != model_size:
            raise ValueError('Model requires square diffraction patterns!')

        model_config = self._create_model_config(model_size)
        inference_config = InferenceConfig(
            model=model_config,
            model_path=Path(),  # not used
            test_data_file=Path(),  # not used
            debug=self._is_developer_mode_enabled,
            output_dir=Path(),  # not used
        )

        # Update global params with new-style config
        update_legacy_dict(ptycho.params.cfg, inference_config)

        # Create RawData
        test_raw_data = create_raw_data(parameters)
        ptycho.probe.set_probe_guess(None, test_raw_data.probeGuess)

        # Group overlapping scan positions
        test_dataset = test_raw_data.generate_grouped_data(
            model_config.N,
            K=self._inference_settings.n_nearest_neighbors.get_value(),
            nsamples=self._inference_settings.n_samples.get_value(),
        )

        # Create PtychoDataContainer
        test_data_container = ptycho.loader.load(
            lambda: test_dataset, test_raw_data.probeGuess, which=None, create_split=False
        )

        # Perform reconstruction
        obj_tensor_full = self._reconstruct_image(test_data_container)

        # Process the reconstructed image (TODO: ptychodus stitches)
        object_out_array = ptycho.tf_helper.reassemble_position(
            obj_tensor_full, test_data_container.global_offsets, M=20
        )

        object_in = parameters.product.object_
        object_out = Object(
            array=numpy.squeeze(object_out_array),
            layer_spacing_m=object_in.layer_spacing_m,
            pixel_geometry=object_in.get_pixel_geometry(),
            center=object_in.get_center(),
        )
        losses: Sequence[LossValue] = list()

        product = Product(
            metadata=parameters.product.metadata,
            probe_positions=parameters.product.probe_positions,
            probes=parameters.product.probes,
            object_=object_out,
            losses=losses,
        )

        yield ReconstructOutput(product)

    def is_model_loaded(self):
        return True  # TODO

    def get_model_file_filter(self) -> str:
        return 'Zipped Archive (*.zip)'

    def load_model_from_file(self, file_path: Path) -> None:
        if file_path.name != self.MODEL_FILE_NAME:
            logger.warning(f"PtychoPINN expects the file name '{self.MODEL_FILE_NAME}'.")

        # TODO model path to/from settings
        self._inference_settings.model_path.set_value(file_path)

        # global config (ptycho.params.cfg) updated during load
        self.__model, self._config = load_inference_bundle(file_path.parent)
        # TODO sync ptycho.params.cfg with settings after load

    def get_training_data_file_filter(self) -> str:
        return 'NumPy Zipped Archive (*.npz)'

    def export_training_data(self, file_path: Path, parameters: ReconstructInput) -> None:
        object_geometry = parameters.product.object_.get_geometry()
        position_x_px: list[float] = list()
        position_y_px: list[float] = list()

        for scan_point in parameters.product.probe_positions:
            object_point = object_geometry.map_coordinates_probe_to_object(scan_point)
            position_x_px.append(object_point.coordinate_x_px)
            position_y_px.append(object_point.coordinate_y_px)

        xcoords = numpy.array(position_x_px)
        ycoords = numpy.array(position_y_px)

        numpy.savez(
            file_path,
            xcoords=xcoords,
            ycoords=ycoords,
            xcoords_start=xcoords,
            ycoords_start=ycoords,
            diff3d=parameters.diffraction_patterns,
            probeGuess=parameters.product.probes.get_probe_no_opr().get_incoherent_mode(0),
            # assume that all patches are from the same object
            objectGuess=parameters.product.object_.get_layer(0),
            scan_index=numpy.zeros(len(parameters.product.probe_positions), dtype=int),
        )

    def train(self, input_path: Path, output_path: Path) -> Iterator[TrainOutput]:
        test_raw_data = RawData.from_file(input_path / 'test_data.npz')  # TODO RawData | None
        train_raw_data = RawData.from_file(input_path / 'train_data.npz')

        model_size = train_raw_data.diff3d.shape[-1]

        if train_raw_data.diff3d.shape[-2] != model_size:
            raise ValueError('Model requires square diffraction patterns!')

        model_config = self._create_model_config(model_size)
        training_config = TrainingConfig(
            model=model_config,
            train_data_file=Path(),  # not used
            test_data_file=None,  # not used
            batch_size=self._training_settings.batch_size.get_value(),
            nepochs=self._training_settings.nepochs.get_value(),
            mae_weight=self._training_settings.mae_weight.get_value(),
            nll_weight=self._training_settings.nll_weight.get_value(),
            realspace_mae_weight=self._training_settings.realspace_mae_weight.get_value(),
            realspace_weight=self._training_settings.realspace_weight.get_value(),
            nphotons=self._training_settings.nphotons.get_value(),  # TODO get from product
            positions_provided=self._training_settings.positions_provided.get_value(),
            probe_trainable=self._training_settings.probe_trainable.get_value(),
            intensity_scale_trainable=self._training_settings.intensity_scale_trainable.get_value(),
            output_dir=Path(),  # not used
        )

        # Update global params with new-style config
        update_legacy_dict(ptycho.params.cfg, training_config)

        from ptycho.workflows.components import run_cdi_example, save_outputs

        recon_amp, recon_phase, train_results = run_cdi_example(
            train_raw_data, test_raw_data, training_config
        )
        model_path = output_path / self.MODEL_FILE_NAME
        ptycho.model_manager.save(output_path)
        save_outputs(recon_amp, recon_phase, train_results, str(output_path))
        self.load_model_from_file(model_path)

        yield TrainOutput()  # TODO yield losses & progress



from PyQt5.QtCore import QRegularExpression
from PyQt5.QtGui import QRegularExpressionValidator
from PyQt5.QtWidgets import QLineEdit, QWidget


class UUIDLineEdit(QLineEdit):
    @staticmethod
    def _create_validator() -> QRegularExpressionValidator:
        hexre = '[0-9A-Fa-f]'
        uuidre = f'{hexre}{{8}}-{hexre}{{4}}-{hexre}{{4}}-{hexre}{{4}}-{hexre}{{12}}'
        return QRegularExpressionValidator(QRegularExpression(uuidre))

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setValidator(UUIDLineEdit._create_validator())



from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Sequence
import concurrent.futures
import logging
import threading

from ptychodus.api.diffraction import (
    BadPixels,
    DiffractionArray,
    SimpleDiffractionArray,
)
from ptychodus.api.io import AssembledDiffractionData

from ..task_manager import BackgroundTask, ForegroundTask, ForegroundTaskManager
from .processor import DiffractionPatternProcessor

logger = logging.getLogger('.'.join(__name__.split('.')[:-1]))


class ArrayAssembler(ABC):
    @abstractmethod
    def _create_array_loader(
        self, array: DiffractionArray, *, process_patterns: bool
    ) -> BackgroundTask:
        pass

    @abstractmethod
    def _assemble_array(
        self,
        array_index: int,
        label: str,
        data: AssembledDiffractionData,
    ) -> None:
        pass


class LoadArray:
    def __init__(
        self,
        array_index: int,
        array: DiffractionArray,
        bad_pixels: BadPixels,
        processor: DiffractionPatternProcessor | None,
        assembler: ArrayAssembler,
    ) -> None:
        super().__init__()
        self._array_index = array_index
        self._array = array
        self._bad_pixels = bad_pixels
        self._processor = processor
        self._assembler = assembler

    def __call__(self) -> ForegroundTask | None:
        label = self._array.get_label()

        try:
            loaded_array = SimpleDiffractionArray(
                label,
                self._array.get_indexes(),
                self._array.get_patterns(),
            )
        except FileNotFoundError:
            logger.warning(f'File not found for "{label}"!')
        else:
            processed_array = (
                loaded_array if self._processor is None else self._processor(loaded_array)
            )
            data = AssembledDiffractionData(
                indexes=processed_array.get_indexes(),
                patterns=processed_array.get_patterns(),
                bad_pixels=self._bad_pixels,
            )
            self._assembler._assemble_array(
                self._array_index,
                label,
                data,
            )

        return None


class LoadAllArrays:
    def __init__(
        self,
        array_seq: Sequence[DiffractionArray],
        assembler: ArrayAssembler,
        foreground_task_manager: ForegroundTaskManager,
    ) -> None:
        super().__init__()
        self._array_seq = array_seq
        self._assembler = assembler
        self._foreground_task_manager = foreground_task_manager
        self._process_patterns = False
        self._finished_event = threading.Event()

    def enable_pattern_processing(self) -> None:
        self._process_patterns = True

    def get_finished_event(self) -> threading.Event:
        return self._finished_event

    def __call__(self) -> None:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            future_list = [
                executor.submit(
                    lambda loader_task: loader_task(),
                    self._assembler._create_array_loader(
                        array,
                        process_patterns=self._process_patterns,
                    ),
                )
                for array in self._array_seq
            ]

            for future in concurrent.futures.as_completed(future_list):
                try:
                    task = future.result()
                except Exception as ex:
                    logger.warning(ex)
                else:
                    if task is not None:
                        self._foreground_task_manager.put_foreground_task(task)

        self._finished_event.set()



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QHBoxLayout,
    QMenu,
    QPlainTextEdit,
    QProgressBar,
    QPushButton,
    QSplitter,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg
from matplotlib.backends.backend_qt import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure


class ProcessingActionsView(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.reconstruct_button = QPushButton()
        self.train_button = QPushButton('Train')
        self.actions_menu = QMenu()
        self.actions_button = QToolButton()

        self.actions_button.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        self.actions_button.setMenu(self.actions_menu)
        self.actions_button.setArrowType(Qt.ArrowType.DownArrow)
        self.actions_button.setStyleSheet('QToolButton::menu-indicator { image: none; }')

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.reconstruct_button)
        layout.addWidget(self.train_button)
        layout.addWidget(self.actions_button)
        self.setLayout(layout)


class ProcessingStatusView(QSplitter):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(Qt.Orientation.Vertical, parent)
        self.figure = Figure()
        self.figure_canvas = FigureCanvasQTAgg(self.figure)
        self.navigation_toolbar = NavigationToolbar(self.figure_canvas, self)
        self.axes = self.figure.add_subplot(111)

        self.text_edit = QPlainTextEdit()
        self.progress_bar = QProgressBar()
        self.stop_button = QPushButton('Stop')

        upper_layout = QVBoxLayout()
        upper_layout.addWidget(self.navigation_toolbar)
        upper_layout.addWidget(self.figure_canvas)

        upper_widget = QWidget()
        upper_widget.setLayout(upper_layout)
        self.addWidget(upper_widget)

        progress_layout = QHBoxLayout()
        progress_layout.addWidget(self.progress_bar)
        progress_layout.addWidget(self.stop_button)

        lower_layout = QVBoxLayout()
        lower_layout.addWidget(self.text_edit)
        lower_layout.addLayout(progress_layout)

        lower_widget = QWidget()
        lower_widget.setLayout(lower_layout)
        self.addWidget(lower_widget)

        self.setStretchFactor(0, 2)
        self.setStretchFactor(1, 0)



from pathlib import Path
from typing import Final


from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.product import Product, ProductFileReader, ProductFileWriter
from ptychodus.api.io import load_product, save_product


class H5ProductFileIO(ProductFileReader, ProductFileWriter):
    SIMPLE_NAME: Final[str] = 'HDF5'
    DISPLAY_NAME: Final[str] = 'Ptychodus Product Files (*.h5 *.hdf5)'

    def read(self, file_path: Path) -> Product:
        return load_product(file_path)

    def write(self, file_path: Path, product: Product) -> None:
        save_product(file_path, product)


def register_plugins(registry: PluginRegistry) -> None:
    h5_product_file_io = H5ProductFileIO()

    registry.register_product_file_reader_with_adapters(
        h5_product_file_io,
        simple_name=H5ProductFileIO.SIMPLE_NAME,
        display_name=H5ProductFileIO.DISPLAY_NAME,
    )
    registry.product_file_writers.register_plugin(
        h5_product_file_io,
        simple_name=H5ProductFileIO.SIMPLE_NAME,
        display_name=H5ProductFileIO.DISPLAY_NAME,
    )



from __future__ import annotations

import numpy

from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider

from .builder import ProbeSequenceBuilder
from .settings import ProbeSettings


class SuperGaussianProbeBuilder(ProbeSequenceBuilder):
    def __init__(self, settings: ProbeSettings) -> None:
        super().__init__(settings, 'super_gaussian')
        self._settings = settings

        self.annular_radius_m = settings.super_gaussian_annular_radius_m.copy()
        self._add_parameter('annular_radius_m', self.annular_radius_m)

        self.fwhm_m = settings.super_gaussian_width_m.copy()
        self._add_parameter('full_width_at_half_maximum_m', self.fwhm_m)

        self.order_parameter = settings.super_gaussian_order_parameter.copy()
        self._add_parameter('order_parameter', self.order_parameter)

    def copy(self) -> SuperGaussianProbeBuilder:
        builder = SuperGaussianProbeBuilder(self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self, geometry_provider: ProbeGeometryProvider) -> ProbeSequence:
        geometry = geometry_provider.get_probe_geometry()
        coords = self.get_transverse_coordinates(geometry)

        Z = (  # noqa: N806
            coords.position_r_m - self.annular_radius_m.get_value()
        ) / self.fwhm_m.get_value()
        ZP = numpy.power(2 * Z, 2 * self.order_parameter.get_value())  # noqa: N806

        return ProbeSequence(
            array=self.normalize(numpy.exp(-numpy.log(2) * ZP) + 0j),
            opr_weights=None,
            pixel_geometry=geometry.get_pixel_geometry(),
        )



from .builder import ObjectBuilder
from .builder_factory import ObjectBuilderFactory
from .dead_leaves import DeadLeavesObjectBuilder
from .item import ObjectRepositoryItem
from .item_factory import ObjectRepositoryItemFactory
from .random import RandomObjectBuilder
from .settings import ObjectSettings
from .stxm import STXMObjectBuilder

__all__ = [
    'DeadLeavesObjectBuilder',
    'ObjectBuilder',
    'ObjectBuilderFactory',
    'ObjectRepositoryItem',
    'ObjectRepositoryItemFactory',
    'ObjectSettings',
    'RandomObjectBuilder',
    'STXMObjectBuilder',
]



from .authorizer import GlobusAuthorizer
from .core import GlobusCore
from .executor import GlobusExecutor
from .settings import GlobusSettings
from .status import GlobusStatusRepository, GlobusStatus

__all__ = [
    'GlobusAuthorizer',
    'GlobusCore',
    'GlobusExecutor',
    'GlobusSettings',
    'GlobusStatus',
    'GlobusStatusRepository',
]



"""2D geometric primitives: points, lines, boxes, intervals, and pixel/image geometry."""

from __future__ import annotations
from dataclasses import dataclass
from decimal import Decimal
from typing import Generic, TypeVar

T = TypeVar('T', int, float, Decimal)


@dataclass(frozen=True)
class AffineTransform:
    """2D affine transformation expressed as a 2x3 matrix; callable as transform(x, y) -> (x', y')."""

    a00: float
    a01: float
    a02: float

    a10: float
    a11: float
    a12: float

    def __call__(self, x: float, y: float) -> tuple[float, float]:
        yp = self.a00 * x + self.a01 * y + self.a02
        xp = self.a10 * x + self.a11 * y + self.a12
        return yp, xp


@dataclass(frozen=True)
class PixelGeometry:
    """Physical dimensions of a single detector or probe pixel in meters."""

    width_m: float
    height_m: float

    def get_area_m2(self) -> float:
        return self.width_m * self.height_m

    def get_aspect_ratio(self) -> float:
        return self.width_m / self.height_m

    def copy(self) -> PixelGeometry:
        return PixelGeometry(
            width_m=float(self.width_m),
            height_m=float(self.height_m),
        )


@dataclass(frozen=True)
class ImageExtent:
    """Integer pixel dimensions (width × height) of an image or detector."""

    width_px: int
    height_px: int

    def get_size(self) -> int:
        """Return the number of pixels in the image."""
        return self.width_px * self.height_px

    def get_shape(self) -> tuple[int, int]:
        """Return the image shape as a (height_px, width_px) tuple."""
        return self.height_px, self.width_px

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ImageExtent):
            return self.get_shape() == other.get_shape()

        return False


@dataclass(frozen=True)
class Point2D:
    """Floating-point 2D point."""

    x: float
    y: float


@dataclass(frozen=True)
class Line2D:
    """2D line segment between two points with linear interpolation."""

    begin: Point2D
    end: Point2D

    def lerp(self, alpha: float) -> Point2D:
        beta = 1 - alpha
        x = beta * self.begin.x + alpha * self.end.x
        y = beta * self.begin.y + alpha * self.end.y
        return Point2D(x, y)


@dataclass(frozen=True)
class Box2D:
    """Axis-aligned 2D bounding box defined by its top-left corner, width, and height."""

    x: float
    y: float
    width: float
    height: float

    @property
    def x_begin(self) -> float:
        return self.x

    @property
    def x_center(self) -> float:
        return self.x + self.width / 2

    @property
    def x_end(self) -> float:
        return self.x + self.width

    @property
    def y_begin(self) -> float:
        return self.y

    @property
    def y_center(self) -> float:
        return self.y + self.height / 2

    @property
    def y_end(self) -> float:
        return self.y + self.height


class Interval(Generic[T]):
    """Closed interval [lower, upper] with clamp, hull, and membership operations."""

    def __init__(self, lower: T, upper: T) -> None:
        self.lower: T = lower
        self.upper: T = upper

    @classmethod
    def create_proper(cls, a: T, b: T) -> Interval[T]:
        if b < a:
            return Interval[T](b, a)
        else:
            return Interval[T](a, b)

    def is_empty(self) -> bool:
        return self.upper < self.lower

    def clamp(self, value: T) -> T:
        return max(self.lower, min(value, self.upper))

    def hull(self, value: Interval[T] | T) -> Interval[T]:
        if isinstance(value, Interval):
            return Interval[T](min(self.lower, value.lower), max(self.upper, value.upper))
        else:
            return Interval[T](min(self.lower, value), max(self.upper, value))

    def get_length(self) -> T:
        return self.upper - self.lower

    def get_midrange(self) -> T:
        total = self.lower + self.upper
        return total // 2 if isinstance(total, int) else total / 2

    def copy(self) -> Interval[T]:
        return Interval[T](self.lower, self.upper)

    def __contains__(self, item: T) -> bool:
        return self.lower <= item and item < self.upper

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.lower}, {self.upper})'



from pathlib import Path
from typing import Final
import logging

import h5py
import numpy

from ptychodus.api.geometry import ImageExtent
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionMetadata,
    SimpleDiffractionDataset,
)
from ptychodus.api.plugins import PluginRegistry

from .h5_diffraction_file import H5DiffractionPatternArray, H5DiffractionFileTreeBuilder

logger = logging.getLogger(__name__)


class NanoMAXDiffractionFileReader(DiffractionFileReader):
    DATA_PATHS: Final[tuple[str, ...]] = (
        '/entry/measurement/eiger1m/frames',
        '/entry/instrument/Eiger/data',
        '/entry/measurement/Eiger/data',
    )

    def __init__(self) -> None:
        self._tree_builder = H5DiffractionFileTreeBuilder()

    def read(self, file_path: Path) -> DiffractionDataset:
        with h5py.File(file_path, 'r') as h5_file:
            contents_tree = self._tree_builder.build(h5_file)
            data: h5py.Dataset | None = None

            for data_path in self.DATA_PATHS:
                data_link = h5_file.get(data_path, getlink=True)

                if data_link is None:
                    logger.debug(f'Data not found at "{data_path}"')
                else:
                    data = h5_file.get(data_path)
                    break

            if data is None:
                raise ValueError('Failed to find data!')
            elif isinstance(data, h5py.Dataset):
                num_patterns, detector_height, detector_width = data.shape

                metadata = DiffractionMetadata(
                    num_patterns_per_array=[num_patterns],
                    pattern_dtype=data.dtype,
                    detector_extent=ImageExtent(detector_width, detector_height),
                    file_path=file_path,
                )

                array = H5DiffractionPatternArray(
                    label=file_path.stem,
                    indexes=numpy.arange(num_patterns),
                    file_path=data.file.filename,
                    data_path=data.name,
                )

                return SimpleDiffractionDataset(metadata, contents_tree, [array])
            else:
                raise ValueError(f'Expected {data.name} to be a dataset; got {type(data)}.')


def register_plugins(registry: PluginRegistry) -> None:
    registry.diffraction_file_readers.register_plugin(
        NanoMAXDiffractionFileReader(),
        simple_name='MAX_IV_NanoMAX',
        display_name='MAX IV NanoMAX Files (*.h5 *.hdf5)',
    )



from abc import abstractmethod
from collections.abc import Iterator

from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.parametric import ParameterGroup
from ptychodus.api.visualization import (
    NumberArrayType,
    RealArrayType,
    VisualizationProduct,
)


class Renderer(ParameterGroup):
    def __init__(self, name: str) -> None:
        super().__init__()
        self._name = self.create_string_parameter('name', name)

    def get_name(self) -> str:
        return self._name.get_value()

    @abstractmethod
    def variants(self) -> Iterator[str]:
        pass

    @abstractmethod
    def get_variant(self) -> str:
        pass

    @abstractmethod
    def set_variant(self, variant: str) -> None:
        pass

    @abstractmethod
    def is_cyclic(self) -> bool:
        pass

    @abstractmethod
    def colorize(self, array: NumberArrayType) -> RealArrayType:
        pass

    @abstractmethod
    def render(
        self, array: NumberArrayType, pixel_geometry: PixelGeometry, *, autoscale_color_axis: bool
    ) -> VisualizationProduct:
        pass



from __future__ import annotations
import logging

from ...model.analysis import FourierRingCorrelator
from ...view.object import FourierRingCorrelationDialog
from .tree_model import ObjectTreeModel

logger = logging.getLogger(__name__)


class FourierRingCorrelationViewController:
    def __init__(self, correlator: FourierRingCorrelator, tree_model: ObjectTreeModel) -> None:
        super().__init__()
        self._correlator = correlator
        self._dialog = FourierRingCorrelationDialog()
        self._dialog.setWindowTitle('Fourier Ring Correlation')
        self._dialog.product1_combo_box.setModel(tree_model)
        self._dialog.product1_combo_box.textActivated.connect(self._redraw_plot)
        self._dialog.product2_combo_box.setModel(tree_model)
        self._dialog.product2_combo_box.textActivated.connect(self._redraw_plot)

    def analyze(self, item_index1: int, item_index2: int) -> None:
        self._dialog.product1_combo_box.setCurrentIndex(item_index1)
        self._dialog.product2_combo_box.setCurrentIndex(item_index2)
        self._redraw_plot()
        self._dialog.open()

    def _redraw_plot(self) -> None:
        current_index1 = self._dialog.product1_combo_box.currentIndex()
        current_index2 = self._dialog.product2_combo_box.currentIndex()

        if current_index1 < 0 or current_index2 < 0:
            logger.warning('Invalid item index for FRC!')
            return

        frc = self._correlator.correlate(current_index1, current_index2)
        plot2d = frc.get_plot()
        axis_x = plot2d.axis_x
        axis_y = plot2d.axis_y

        ax = self._dialog.axes
        ax.clear()
        ax.set_xlabel(axis_x.label)
        ax.set_ylabel(axis_y.label)
        ax.grid(True)

        if len(axis_x.series) == 1:
            sx = axis_x.series[0]

            for sy in axis_y.series:
                ax.plot(sx.values, sy.values, '.-', label=sy.label, linewidth=1.5)
        else:
            logger.warning('Failed to broadcast plot series!')

        if len(axis_x.series) > 1:
            ax.legend(loc='upper right')

        self._dialog.figure_canvas.draw()



from __future__ import annotations
from decimal import Decimal

import numpy

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QHBoxLayout, QLabel, QSlider, QWidget

from ptychodus.api.geometry import Interval


class DecimalSlider(QWidget):
    value_changed = pyqtSignal(Decimal)

    def __init__(self, slider: QSlider, parent: QWidget | None) -> None:
        super().__init__(parent)
        self._slider = slider
        self._label = QLabel()
        self._value = Decimal()
        self._minimum = Decimal()
        self._maximum = Decimal()

    @classmethod
    def create_instance(
        cls,
        orientation: Qt.Orientation,
        parent: QWidget | None = None,
        *,
        num_ticks: int = 1000,
    ) -> DecimalSlider:
        slider = QSlider(orientation)
        slider.setRange(0, num_ticks)
        slider.setTickPosition(QSlider.TickPosition.TicksBelow)
        slider.setTickInterval(100)

        widget = cls(slider, parent)
        slider.valueChanged.connect(lambda value: widget._set_value_from_slider())
        widget.set_value_and_range(Decimal(1) / 2, Interval[Decimal](Decimal(0), Decimal(1)))

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widget._slider)
        layout.addWidget(widget._label)
        widget.setLayout(layout)

        return widget

    def get_value(self) -> Decimal:
        return self._value

    def set_value(self, value: Decimal) -> None:
        if self._set_value_to_slider(value):
            self._emit_value_changed()

    def set_value_and_range(
        self,
        value: Decimal,
        range_: Interval[Decimal],
        block_value_changed_signal: bool = False,
    ) -> None:
        should_emit = False

        if range_.upper <= range_.lower:
            raise ValueError(f'maximum <= minimum ({range_.upper} <= {range_.lower})')

        if range_.lower != self._minimum:
            self._minimum = range_.lower
            should_emit = True

        if range_.upper != self._maximum:
            self._maximum = range_.upper
            should_emit = True

        if self._set_value_to_slider(value):
            should_emit = True

        if not block_value_changed_signal and should_emit:
            self._emit_value_changed()

    def _set_value_from_slider(self) -> None:
        upper = Decimal(self._slider.value() - self._slider.minimum())
        lower = Decimal(self._slider.maximum() - self._slider.minimum())
        alpha = upper / lower
        value = (1 - alpha) * self._minimum + alpha * self._maximum

        if value != self._value:
            self._value = value
            self._update_label()
            self._emit_value_changed()

    def _set_value_to_slider(self, value: Decimal) -> bool:
        should_emit = False

        alpha = (Decimal(value) - self._minimum) / (self._maximum - self._minimum)
        ivaluef = (1 - alpha) * self._slider.minimum() + alpha * self._slider.maximum()
        ivalue = int(numpy.rint(float(ivaluef)))

        if value < self._minimum:
            value = self._minimum
            ivalue = self._slider.minimum()
        elif value > self._maximum:
            value = self._maximum
            ivalue = self._slider.maximum()

        self._slider.blockSignals(True)
        self._slider.setValue(ivalue)
        self._slider.blockSignals(False)

        if value != self._value:
            self._value = Decimal(value)
            self._update_label()
            should_emit = True

        return should_emit

    def _update_label(self) -> None:
        self._label.setText(f'{self._value:.3f}')

    def _emit_value_changed(self) -> None:
        self.value_changed.emit(self._value)



from .controller import VisualizationController
from .parameters import VisualizationParametersController
from .widget import VisualizationWidgetController

__all__ = [
    'VisualizationController',
    'VisualizationParametersController',
    'VisualizationWidgetController',
]



from collections.abc import Iterator
from importlib.metadata import version
import logging

from ptychodus.api.reconstructor import (
    NullReconstructor,
    Reconstructor,
    ReconstructorLibrary,
)
from ptychodus.api.settings import SettingsRegistry

from ..diffraction import PatternSizer
from .device import PtyChiDeviceRepository
from .enums import PtyChiEnumerators
from .settings import (
    PtyChiAutodiffSettings,
    PtyChiDMSettings,
    PtyChiLSQMLSettings,
    PtyChiOPRSettings,
    PtyChiObjectSettings,
    PtyChiPIESettings,
    PtyChiProbePositionSettings,
    PtyChiProbeSettings,
    PtyChiSettings,
)

logger = logging.getLogger(__name__)


class PtyChiReconstructorLibrary(ReconstructorLibrary):
    def __init__(
        self,
        settings_registry: SettingsRegistry,
        pattern_sizer: PatternSizer,
        is_developer_mode_enabled: bool,
    ) -> None:
        super().__init__('ptychi')
        self.autodiff_settings = PtyChiAutodiffSettings(settings_registry)
        self.dm_settings = PtyChiDMSettings(settings_registry)
        self.lsqml_settings = PtyChiLSQMLSettings(settings_registry)
        self.object_settings = PtyChiObjectSettings(settings_registry)
        self.opr_settings = PtyChiOPRSettings(settings_registry)
        self.pie_settings = PtyChiPIESettings(settings_registry)
        self.probe_position_settings = PtyChiProbePositionSettings(settings_registry)
        self.probe_settings = PtyChiProbeSettings(settings_registry)
        self.settings = PtyChiSettings(settings_registry)

        self.enumerators = PtyChiEnumerators()
        self.device_repository = PtyChiDeviceRepository(
            is_developer_mode_enabled=is_developer_mode_enabled
        )
        self.reconstructor_list: list[Reconstructor] = list()

        try:
            from .autodiff import AutodiffReconstructor
            from .dm import DMReconstructor
            from .epie import EPIEReconstructor
            from .helper import PtyChiOptionsHelper
            from .lsqml import LSQMLReconstructor
            from .pie import PIEReconstructor
            from .rpie import RPIEReconstructor
        except ModuleNotFoundError:
            logger.info('pty-chi not found.')

            if is_developer_mode_enabled:
                for reconstructor in ('DM', 'PIE', 'ePIE', 'rPIE', 'LSQML', 'Autodiff'):
                    self.reconstructor_list.append(NullReconstructor(reconstructor))
        else:
            ptychi_version = version('ptychi')
            logger.info(f'Pty-Chi {ptychi_version}')

            options_helper = PtyChiOptionsHelper(
                self.settings,
                self.object_settings,
                self.probe_settings,
                self.probe_position_settings,
                self.opr_settings,
                pattern_sizer,
            )
            self.reconstructor_list.append(DMReconstructor(options_helper, self.dm_settings))
            self.reconstructor_list.append(PIEReconstructor(options_helper, self.pie_settings))
            self.reconstructor_list.append(EPIEReconstructor(options_helper, self.pie_settings))
            self.reconstructor_list.append(RPIEReconstructor(options_helper, self.pie_settings))
            self.reconstructor_list.append(LSQMLReconstructor(options_helper, self.lsqml_settings))
            self.reconstructor_list.append(
                AutodiffReconstructor(options_helper, self.autodiff_settings)
            )

    @property
    def name(self) -> str:
        return 'pty-chi'

    def __iter__(self) -> Iterator[Reconstructor]:
        return iter(self.reconstructor_list)



from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Iterable
import logging

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QFormLayout,
    QLabel,
    QScrollArea,
    QStackedWidget,
    QWidget,
)

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.reconstructor import TrainableReconstructor

from ...model.globus import GlobusCore
from ...model.processing import ProcessingAPI, ProcessingAlgorithmParameter
from ...model.product import ProductRepository
from ...view.processing import ProcessingActionsView, ProcessingStatusView
from ...view.widgets import ExceptionDialog
from ..data import FileDialogFactory
from ..helpers import connect_triggered_signal
from ..parametric import ComboBoxParameterViewController
from .parameters import (
    ProcessingStatusController,
    ProductParameterViewController,
    ComputeParameterViewController,
)

logger = logging.getLogger(__name__)


class ReconstructorViewControllerFactory(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def create_view_controller(self, reconstructor_name: str) -> QWidget:
        pass


class ProcessingController(Observer):
    def __init__(
        self,
        algorithm_parameter: ProcessingAlgorithmParameter,
        processing_api: ProcessingAPI,
        product_repository: ProductRepository,
        globus: GlobusCore,
        view: QWidget,
        status_view: ProcessingStatusView,
        file_dialog_factory: FileDialogFactory,
        view_controller_factories: Iterable[ReconstructorViewControllerFactory],
    ) -> None:
        super().__init__()
        self._algorithm_parameter = algorithm_parameter
        self._processing_api = processing_api
        self._product_repository = product_repository
        self._globus = globus
        self._view = view
        self._status_view = status_view
        self._file_dialog_factory = file_dialog_factory

        self._stacked_widget = QStackedWidget()
        self._populate_stacked_widget(view_controller_factories)
        stacked_widget_layout = self._stacked_widget.layout()

        if stacked_widget_layout is not None:
            stacked_widget_layout.setContentsMargins(0, 0, 0, 0)

        self._scroll_area = QScrollArea()
        self._scroll_area.setWidgetResizable(True)
        self._scroll_area.setWidget(self._stacked_widget)

        self._algorithm_view_controller = ComboBoxParameterViewController(
            algorithm_parameter, algorithm_parameter.available_reconstructors()
        )
        self._algorithm_view_controller.get_widget().currentIndexChanged.connect(
            self._stacked_widget.setCurrentIndex
        )
        self._status_controller = ProcessingStatusController(
            product_repository, processing_api.get_progress_monitor(), status_view
        )
        self._product_view_controller = ProductParameterViewController(
            product_repository, self._status_controller
        )
        self._compute_view_controller = ComputeParameterViewController()
        self._actions_view = ProcessingActionsView()

        layout = QFormLayout()
        layout.addRow('Algorithm:', self._algorithm_view_controller.get_widget())
        layout.addRow('Product:', self._product_view_controller.get_widget())

        if globus.is_supported:
            layout.addRow('Compute:', self._compute_view_controller.get_widget())

        layout.addRow('Action:', self._actions_view)
        layout.addRow(self._scroll_area)
        self._view.setLayout(layout)

        self._actions_view.reconstruct_button.clicked.connect(self._reconstruct)
        self._actions_view.train_button.clicked.connect(self._train)

        self._reconstruct_split_action = self._actions_view.actions_menu.addAction(
            'Reconstruct Split'
        )
        self._reconstruct_split_action.setEnabled(False)  # TODO
        connect_triggered_signal(self._reconstruct_split_action, self._reconstruct_split)
        self._load_model_action = self._actions_view.actions_menu.addAction('Load Model...')
        connect_triggered_signal(self._load_model_action, self._load_model)
        self._export_training_data_action = self._actions_view.actions_menu.addAction(
            'Export Training Data...'
        )
        connect_triggered_signal(self._export_training_data_action, self._export_training_data)

        self._sync_model_to_view()
        algorithm_parameter.add_observer(self)

    def _populate_stacked_widget(
        self, view_controller_factories: Iterable[ReconstructorViewControllerFactory]
    ) -> None:
        vcf_map = {vcf.name.casefold(): vcf for vcf in view_controller_factories}

        for library, reconstructor in self._processing_api.available_reconstructors_parts():
            try:
                vcf = vcf_map[library]
            except KeyError:
                label = QLabel(f'{library} not found!')
                logger.warning(label.text())
                label.setAlignment(Qt.AlignmentFlag.AlignCenter)
                widget: QWidget = label
            else:
                widget = vcf.create_view_controller(reconstructor)

            self._stacked_widget.addWidget(widget)

    def _reconstruct(self) -> None:
        input_product_index = self._product_view_controller.get_widget().currentIndex()

        if input_product_index < 0:
            return

        if self._compute_view_controller.is_computing_local():
            try:
                output_product_index = self._processing_api.reconstruct(input_product_index)
            except Exception as exc:
                logger.exception(exc)
                ExceptionDialog.show_exception('Reconstruct Local', exc)
            else:
                self._product_view_controller.get_widget().setCurrentIndex(output_product_index)
        else:
            try:
                self._globus.executor.reconstruct(input_product_index)
            except Exception as exc:
                logger.exception(exc)
                ExceptionDialog.show_exception('Reconstruct Remote', exc)

    def _train(self) -> None:
        product_index = self._product_view_controller.get_widget().currentIndex()

        if product_index < 0:
            return

        data_path = self._file_dialog_factory.get_existing_directory_path(
            self._view, 'Choose Training Data Directory'
        )

        if not data_path:
            return

        if self._compute_view_controller.is_computing_local():
            try:
                self._processing_api.train(product_index, data_path, data_path)
            except Exception as exc:
                logger.exception(exc)
                ExceptionDialog.show_exception('Train Local', exc)
        else:
            try:
                self._globus.executor.train(product_index)
            except Exception as exc:
                logger.exception(exc)
                ExceptionDialog.show_exception('Train Remote', exc)

    def _reconstruct_split(self) -> None:
        pass  # TODO

    def _load_model(self) -> None:
        name_filter = self._processing_api.get_model_file_filter()
        file_path, name_filter = self._file_dialog_factory.get_open_file_path(
            self._view, 'Load Model', name_filters=[name_filter], selected_name_filter=name_filter
        )

        if not file_path:
            return

        try:
            self._processing_api.load_model_from_file(file_path)
        except Exception as exc:
            logger.exception(exc)
            ExceptionDialog.show_exception('Load Model', exc)

    def _export_training_data(self) -> None:
        input_product_index = self._product_view_controller.get_widget().currentIndex()

        if input_product_index < 0:
            return

        name_filter = self._processing_api.get_training_data_file_filter()
        file_path, _ = self._file_dialog_factory.get_save_file_path(
            self._view,
            'Export Training Data',
            name_filters=[name_filter],
            selected_name_filter=name_filter,
        )

        if not file_path:
            return

        try:
            self._processing_api.export_training_data(file_path, input_product_index)
        except Exception as exc:
            logger.exception(exc)
            ExceptionDialog.show_exception('Export Training Data', exc)

    def _sync_model_to_view(self) -> None:
        reconstructor = self._algorithm_parameter.get_current_reconstructor()
        is_trainable = False

        if isinstance(reconstructor, TrainableReconstructor):
            is_trainable = True
            is_model_loaded = reconstructor.is_model_loaded()
            self._actions_view.reconstruct_button.setText('Infer')
            self._actions_view.reconstruct_button.setEnabled(is_model_loaded)
        else:
            self._actions_view.reconstruct_button.setText('Reconstruct')
            self._actions_view.reconstruct_button.setEnabled(True)

        self._actions_view.train_button.setVisible(is_trainable)
        self._load_model_action.setVisible(is_trainable)
        self._export_training_data_action.setVisible(is_trainable)

    def _update(self, observable: Observable) -> None:
        if observable is self._algorithm_parameter:
            self._sync_model_to_view()



from pathlib import Path

import numpy

from ptychodus.api.diffraction import BadPixels, BadPixelsFileReader
from ptychodus.api.plugins import PluginRegistry


class NPYBadPixelsFileReader(BadPixelsFileReader):
    def read(self, file_path: Path) -> BadPixels:
        return numpy.load(file_path)


class NPYGoodPixelsFileReader(BadPixelsFileReader):
    def read(self, file_path: Path) -> BadPixels:
        return numpy.logical_not(numpy.load(file_path))


def register_plugins(registry: PluginRegistry) -> None:
    registry.bad_pixels_file_readers.register_plugin(
        NPYBadPixelsFileReader(),
        simple_name='NPY_Bad_Pixels',
        display_name='NumPy Bad Pixel Files (*.npy)',
    )
    registry.bad_pixels_file_readers.register_plugin(
        NPYGoodPixelsFileReader(),
        simple_name='NPY_Good_Pixels',
        display_name='NumPy Good Pixel Files (*.npy)',
    )



from __future__ import annotations

import numpy

from ptychodus.api.probe_positions import ProbePositionSequence, ProbePosition

from .builder import ProbePositionsBuilder
from .settings import ProbePositionsSettings


class ConcentricProbePositionsBuilder(ProbePositionsBuilder):
    """https://doi.org/10.1088/1367-2630/12/3/035017"""

    def __init__(self, settings: ProbePositionsSettings) -> None:
        super().__init__(settings, 'concentric')
        self._settings = settings

        self.radial_step_size_m = settings.radial_step_size_m.copy()
        self._add_parameter('radial_step_size_m', self.radial_step_size_m)

        self.num_shells = settings.num_shells.copy()
        self._add_parameter('num_shells', self.num_shells)

        self.num_points_1st_shell = settings.num_points_in_first_shell.copy()
        self._add_parameter('num_points_1st_shell', self.num_points_1st_shell)

    def copy(self) -> ConcentricProbePositionsBuilder:
        builder = ConcentricProbePositionsBuilder(self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    @property
    def _num_points(self) -> int:
        num_shells = self.num_shells.get_value()
        triangle = (num_shells * (num_shells + 1)) // 2
        return triangle * self.num_points_1st_shell.get_value()

    def build(self) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()

        for index in range(self._num_points):
            triangle = index // self.num_points_1st_shell.get_value()
            shell_index = int((1 + numpy.sqrt(1 + 8 * triangle)) / 2) - 1  # see OEIS A002024
            shell_triangle = (shell_index * (shell_index + 1)) // 2
            first_index_in_shell = self.num_points_1st_shell.get_value() * shell_triangle
            point_index_in_shell = index - first_index_in_shell

            radius_m = self.radial_step_size_m.get_value() * (shell_index + 1)
            num_points_in_shell = self.num_points_1st_shell.get_value() * (shell_index + 1)
            theta_rad = 2 * numpy.pi * point_index_in_shell / num_points_in_shell

            point = ProbePosition(
                index=index,
                coordinate_x_m=radius_m * numpy.cos(theta_rad),
                coordinate_y_m=radius_m * numpy.sin(theta_rad),
            )
            point_list.append(point)

        return ProbePositionSequence(point_list)



import logging

from ptychodus.api.observer import Observable

logger = logging.getLogger(__name__)


class GlobusAuthorizer(Observable):
    def __init__(self) -> None:
        super().__init__()
        self._authorize_url = ''
        self._authorize_code = ''

    @property
    def _has_authorize_url(self) -> bool:
        auth_url_empty = not self._authorize_url
        return not auth_url_empty

    @property
    def has_authorize_code(self) -> bool:
        auth_code_empty = not self._authorize_code
        return not auth_code_empty

    @property
    def needs_authorize_code(self) -> bool:
        return self._has_authorize_url and not self.has_authorize_code

    def _authorize(self, url: str) -> None:
        logger.info(f'Authorize at {url}')
        self._authorize_url = url
        self._authorize_code = ''
        self.notify_observers()

    def get_authorize_url(self) -> str:
        return self._authorize_url

    def set_code_from_authorize_url(self, code: str) -> None:
        logger.info('Received authorization code.')
        self._authorize_code = code
        self.notify_observers()

    def get_authorize_code(self) -> str:
        return self._authorize_code

    def cancel_authorization(self) -> None:
        logger.info('Canceled authorization')
        self._authorize_url = ''
        self._authorize_code = ''
        self.notify_observers()


class AuthorizeWithGlobus:
    def __init__(self, authorizer: GlobusAuthorizer, authorize_url: str) -> None:
        self._authorizer = authorizer
        self._authorize_url = authorize_url

    def __call__(self) -> None:
        self._authorizer._authorize(self._authorize_url)



#!/usr/bin/env python

from __future__ import annotations

import argparse
import importlib.util
import logging
import shutil
import sys
from pathlib import Path

import numpy

from ptychodus.model import ModelCore

logger = logging.getLogger(__name__)

DEFAULT_DATASET = Path('../PtychoPINN/datasets/Run1084_recon3_postPC_shrunk_3.npz')
REQUIRED_DATASET_KEYS = {
    'diffraction',
    'xcoords_start',
    'ycoords_start',
    'probeGuess',
    'objectGuess',
}
REQUIRED_TRAIN_KEYS = {
    'xcoords',
    'ycoords',
    'diff3d',
    'probeGuess',
    'objectGuess',
}


def _require_module(module_name: str) -> None:
    if importlib.util.find_spec(module_name) is None:
        raise RuntimeError(f'Missing required module "{module_name}".')


def _validate_npz_keys(file_path: Path, required_keys: set[str], label: str) -> None:
    with numpy.load(file_path) as npz_file:
        keys = set(npz_file.files)

    missing = sorted(required_keys - keys)

    if missing:
        missing_keys = ', '.join(missing)
        raise RuntimeError(f'{label} missing keys: {missing_keys}')


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog='ptychodus-ptychopinn-tf-test',
        description='Run the PtychoPINN TF train/load/infer workflow via ptychodus.',
    )
    parser.add_argument(
        '--dataset',
        type=Path,
        default=DEFAULT_DATASET,
        help='Path to the SLAC NPZ dataset.',
    )
    parser.add_argument(
        '--output-dir',
        type=Path,
        required=True,
        help='Directory to write training data and reconstruction outputs.',
    )
    parser.add_argument(
        '--settings',
        type=Path,
        help='Optional settings file for ptychodus.',
    )
    parser.add_argument(
        '--log-level',
        type=int,
        default=logging.INFO,
        help='Python logging level.',
    )
    parser.add_argument(
        '--detector-distance-m',
        type=float,
        default=0.75,
        help='Detector distance (meters).',
    )
    parser.add_argument(
        '--probe-energy-ev',
        type=float,
        default=8000.0,
        help='Probe energy (eV).',
    )
    parser.add_argument(
        '--exposure-time-s',
        type=float,
        default=0.1,
        help='Exposure time (seconds).',
    )
    parser.add_argument(
        '--nepochs',
        type=int,
        default=50,
        help='Number of training epochs.',
    )
    parser.add_argument(
        '--batch-size',
        type=int,
        default=16,
        help='Training batch size.',
    )
    parser.add_argument(
        '--gridsize',
        type=int,
        default=1,
        help='PtychoPINN gridsize setting.',
    )
    parser.add_argument(
        '--test-samples',
        type=int,
        default=512,
        help='Number of inference samples to use.',
    )
    parser.add_argument(
        '--reconstructor',
        default='PtychoPINN/PINN',
        help='Reconstructor name to use.',
    )

    return parser.parse_args()


def main() -> int:
    args = _parse_args()

    logging.basicConfig(
        format='%(asctime)s [%(levelname)s] %(name)s: %(message)s',
        stream=sys.stdout,
        encoding='utf-8',
        level=args.log_level,
    )

    try:
        _require_module('ptycho')
    except RuntimeError as exc:
        logger.error(exc)
        return 1

    dataset_path = args.dataset
    if not dataset_path.is_file():
        logger.error(f'Dataset not found: {dataset_path}')
        return 1

    try:
        _validate_npz_keys(dataset_path, REQUIRED_DATASET_KEYS, 'dataset')
    except RuntimeError as exc:
        logger.error(exc)
        return 1

    output_dir = args.output_dir
    train_dir = output_dir / 'train'
    model_out_dir = output_dir / 'model_out'
    output_dir.mkdir(parents=True, exist_ok=True)
    train_dir.mkdir(parents=True, exist_ok=True)
    model_out_dir.mkdir(parents=True, exist_ok=True)

    try:
        with ModelCore(args.settings, log_level=args.log_level) as model:
            workflow_diffraction_api = model.workflow_api.load_diffraction_data(
                dataset_path,
                file_type='SLAC_NPZ',
                process_patterns=True,
                block=True,
            )
            max_pattern_counts = (
                workflow_diffraction_api.get_assembled_data().get_pattern_counts().max()
            )
            input_product_api = model.workflow_api.create_product(
                name='Run1084_recon3_postPC_shrunk_3',
                detector_distance_m=args.detector_distance_m,
                probe_energy_eV=args.probe_energy_ev,
                probe_photon_count=max_pattern_counts,
                exposure_time_s=args.exposure_time_s,
            )
            input_product_api.load_probe_positions(dataset_path, file_type='SLAC_NPZ')
            input_product_api.load_probe(dataset_path, file_type='SLAC_NPZ')
            input_product_api.load_object(dataset_path, file_type='SLAC_NPZ')

            # TODO: find better way to change settings via Python API
            model_settings = model.ptychopinn_reconstructor_library.model_settings
            model_settings.gridsize.set_value(args.gridsize)

            train_data_path = train_dir / 'train_data.npz'
            input_product_api.export_training_data(train_data_path, algorithm=args.reconstructor)
            test_data_path = train_dir / 'test_data.npz'
            shutil.copyfile(train_data_path, test_data_path)

            _validate_npz_keys(train_data_path, REQUIRED_TRAIN_KEYS, 'train_data.npz')
            _validate_npz_keys(test_data_path, REQUIRED_TRAIN_KEYS, 'test_data.npz')

            # TODO: find better way to change settings via Python API
            training_settings = model.ptychopinn_reconstructor_library.training_settings
            training_settings.nepochs.set_value(args.nepochs)
            training_settings.batch_size.set_value(args.batch_size)

            # TODO: find better way to change settings via Python API
            inference_settings = model.ptychopinn_reconstructor_library.inference_settings
            inference_settings.n_samples.set_value(args.test_samples)

            input_product_api.train_reconstructor_local(
                train_dir, model_out_dir, algorithm=args.reconstructor
            )

            model_file = model_out_dir / 'wts.h5.zip'

            if not model_file.is_file():
                logger.error(f'Model file not found: {model_file}')
                return 1

            output_product_api = input_product_api.reconstruct_local(
                algorithm=args.reconstructor, block=True
            )

            recon_product_path = output_dir / 'recon_product.h5'
            output_product_api.save_product(recon_product_path, file_type='HDF5')
            output_product = output_product_api.get_product()
            object_array = output_product.object_.get_layer(0)

            if object_array.size == 0:
                logger.error('Reconstruction produced an empty object array.')
                return 1

            numpy.save(output_dir / 'recon_object.npy', object_array)
    except Exception:
        logger.exception('PtychoPINN TF workflow failed.')
        return 1

    return 0


if __name__ == '__main__':
    sys.exit(main())



from abc import ABC, abstractmethod

from skimage.restoration import unwrap_phase
import numpy

from ptychodus.api.common import NumberArrayType, RealArrayType


class DataArrayComponent(ABC):
    def __init__(self, name: str, *, is_cyclic: bool) -> None:
        self._name = name
        self._is_cyclic = is_cyclic

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_cyclic(self) -> bool:
        return self._is_cyclic

    @abstractmethod
    def calculate(self, array: NumberArrayType) -> RealArrayType:
        pass


class RealArrayComponent(DataArrayComponent):
    def __init__(self) -> None:
        super().__init__('real', is_cyclic=False)

    def calculate(self, array: NumberArrayType) -> RealArrayType:
        return numpy.real(array).astype(numpy.single)


class ImaginaryArrayComponent(DataArrayComponent):
    def __init__(self) -> None:
        super().__init__('imaginary', is_cyclic=False)

    def calculate(self, array: NumberArrayType) -> RealArrayType:
        return numpy.imag(array).astype(numpy.single)


class AmplitudeArrayComponent(DataArrayComponent):
    def __init__(self) -> None:
        super().__init__('amplitude', is_cyclic=False)

    def calculate(self, array: NumberArrayType) -> RealArrayType:
        return numpy.absolute(array).astype(numpy.single)


class PhaseInRadiansArrayComponent(DataArrayComponent):
    def __init__(self) -> None:
        super().__init__('phase', is_cyclic=True)

    def calculate(self, array: NumberArrayType) -> RealArrayType:
        return numpy.angle(array).astype(numpy.single)  # type: ignore


class UnwrappedPhaseInRadiansArrayComponent(DataArrayComponent):
    def __init__(self) -> None:
        super().__init__('unwrapped_phase', is_cyclic=False)

    def calculate(self, array: NumberArrayType) -> RealArrayType:
        phase_rad = numpy.angle(array).astype(numpy.single)  # type: ignore
        return unwrap_phase(phase_rad)



from __future__ import annotations
from collections.abc import Sequence
from typing import Any
import logging

from PyQt5.QtCore import (
    QAbstractTableModel,
    QModelIndex,
    QObject,
    QSortFilterProxyModel,
    Qt,
)
from PyQt5.QtGui import QBrush
from PyQt5.QtWidgets import QAbstractItemView, QAction

from ptychodus.api.common import BYTES_PER_MEGABYTE
from ptychodus.api.product import LossValue

from ...model.product import (
    ProductAPI,
    ProductRepository,
    ProductRepositoryItem,
    ProductRepositoryObserver,
)
from ...model.diffraction import DiffractionAPI
from ...model.product.metadata import MetadataRepositoryItem
from ...model.product.object import ObjectRepositoryItem
from ...model.product.probe import ProbeRepositoryItem
from ...model.product.probe_positions import ProbePositionsRepositoryItem
from ...view.product import ProductView
from ...view.widgets import ExceptionDialog
from ..data import FileDialogFactory
from ..helpers import (
    connect_current_changed_signal,
    connect_triggered_signal,
    create_brush_for_editable_cell,
)
from .editor import ProductEditorViewController

logger = logging.getLogger(__name__)


class ProductRepositoryTableModel(QAbstractTableModel):
    def __init__(
        self,
        repository: ProductRepository,
        editable_item_brush: QBrush,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(parent)
        self._repository = repository
        self._editable_item_brush = editable_item_brush
        self._header = [
            'Name',
            'Detector-Object\nDistance [m]',
            'Probe Energy\n[keV]',
            'Probe Photon\nCount',
            'Pixel Width\n[nm]',
            'Pixel Height\n[nm]',
            'Size\n[MB]',
        ]

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        value = super().flags(index)

        if index.isValid() and index.column() < 4:
            value |= Qt.ItemFlag.ItemIsEditable

        return value

    def headerData(  # noqa: N802
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return self._header[section]

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if index.isValid():
            try:
                item = self._repository[index.row()]
            except IndexError as err:
                logger.exception(err)
                return None

            metadata_item = item.get_metadata_item()
            geometry = item.get_geometry()

            if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
                match index.column():
                    case 0:
                        return metadata_item.name.get_value()
                    case 1:
                        return f'{metadata_item.detector_distance_m.get_value():.4g}'
                    case 2:
                        return f'{metadata_item.probe_energy_eV.get_value() / 1e3:.4g}'
                    case 3:
                        return f'{metadata_item.probe_photon_count.get_value():.4g}'
                    case 4:
                        return f'{geometry.object_plane_pixel_width_m * 1e9:.4g}'
                    case 5:
                        return f'{geometry.object_plane_pixel_height_m * 1e9:.4g}'
                    case 6:
                        product = item.get_product()
                        return f'{product.nbytes / BYTES_PER_MEGABYTE:.2f}'
            elif role == Qt.ItemDataRole.BackgroundRole:
                if index.flags() & Qt.ItemFlag.ItemIsEditable:
                    return self._editable_item_brush

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole) -> bool:  # noqa: N802
        if index.isValid() and role == Qt.ItemDataRole.EditRole:
            try:
                item = self._repository[index.row()]
            except IndexError as err:
                logger.exception(err)
                return False

            metadata_item = item.get_metadata_item()

            if index.column() == 0:
                metadata_item.name.set_value(str(value))
                return True
            elif index.column() == 1:
                try:
                    distance_m = float(value)
                except ValueError:
                    return False

                metadata_item.detector_distance_m.set_value(distance_m)
                return True
            elif index.column() == 2:
                try:
                    energy_keV = float(value)  # noqa: N806
                except ValueError:
                    return False

                metadata_item.probe_energy_eV.set_value(energy_keV * 1e3)
                return True
            elif index.column() == 3:
                try:
                    photon_count = float(value)
                except ValueError:
                    return False

                metadata_item.probe_photon_count.set_value(photon_count)
                return True

        return False

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._repository)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._header)


class ProductController(ProductRepositoryObserver):
    def __init__(
        self,
        diffraction_api: DiffractionAPI,
        repository: ProductRepository,
        api: ProductAPI,
        view: ProductView,
        file_dialog_factory: FileDialogFactory,
        duplicate_action: QAction,
        table_model: ProductRepositoryTableModel,
        table_proxy_model: QSortFilterProxyModel,
    ) -> None:
        super().__init__()
        self._diffraction_api = diffraction_api
        self._repository = repository
        self._api = api
        self._view = view
        self._file_dialog_factory = file_dialog_factory
        self._duplicate_action = duplicate_action
        self._table_model = table_model
        self._table_proxy_model = table_proxy_model

    @classmethod
    def create_instance(
        cls,
        diffraction_api: DiffractionAPI,
        repository: ProductRepository,
        api: ProductAPI,
        view: ProductView,
        file_dialog_factory: FileDialogFactory,
    ) -> ProductController:
        open_file_action = view.button_box.insert_menu.addAction('Open File...')
        create_new_action = view.button_box.insert_menu.addAction('Create New')
        duplicate_action = view.button_box.insert_menu.addAction('Duplicate')
        save_file_action = view.button_box.save_menu.addAction('Save File...')
        sync_to_settings_action = view.button_box.save_menu.addAction('Sync To Settings')

        editable_item_brush = create_brush_for_editable_cell(view.table_view)
        table_model = ProductRepositoryTableModel(repository, editable_item_brush)

        table_proxy_model = QSortFilterProxyModel()
        table_proxy_model.setSourceModel(table_model)

        controller = cls(
            diffraction_api,
            repository,
            api,
            view,
            file_dialog_factory,
            duplicate_action,
            table_model,
            table_proxy_model,
        )
        repository.add_observer(controller)
        controller._update_info_text()

        view.table_view.setModel(table_proxy_model)
        view.table_view.setSortingEnabled(True)
        view.table_view.sortByColumn(0, Qt.SortOrder.AscendingOrder)
        vertical_header = view.table_view.verticalHeader()

        if vertical_header is not None:
            vertical_header.hide()

        view.table_view.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        connect_current_changed_signal(view.table_view, controller._update_enabled_buttons)
        controller._update_enabled_buttons(QModelIndex(), QModelIndex())

        connect_triggered_signal(open_file_action, controller._open_product_from_file)
        connect_triggered_signal(create_new_action, controller._create_new_product)
        connect_triggered_signal(duplicate_action, controller._duplicate_current_product)
        connect_triggered_signal(save_file_action, controller._save_current_product_to_file)
        connect_triggered_signal(
            sync_to_settings_action, controller._sync_current_product_to_settings
        )

        view.button_box.edit_button.clicked.connect(controller._edit_current_product)
        view.button_box.remove_button.clicked.connect(controller._remove_current_product)

        return controller

    @property
    def table_model(self) -> QAbstractTableModel:
        return self._table_model

    def _get_current_item_index(self) -> int:
        proxy_index = self._view.table_view.currentIndex()

        if proxy_index.isValid():
            model_index = self._table_proxy_model.mapToSource(proxy_index)
            return model_index.row()

        logger.warning('No current index!')
        return -1

    def _open_product_from_file(self) -> None:
        file_path, name_filter = self._file_dialog_factory.get_open_file_path(
            self._view,
            'Open Product',
            name_filters=[nf for nf in self._api.get_open_file_filters()],
            selected_name_filter=self._api.get_open_file_filter(),
        )

        if file_path:
            try:
                self._api.open_product(file_path, file_type=name_filter)
            except Exception as err:
                logger.exception(err)
                ExceptionDialog.show_exception('File Reader', err)

    def _create_new_product(self) -> None:
        self._api.insert_new_product()

    def _save_current_product_to_file(self) -> None:
        current = self._table_proxy_model.mapToSource(self._view.table_view.currentIndex())

        if current.isValid():
            file_path, name_filter = self._file_dialog_factory.get_save_file_path(
                self._view,
                'Save Product',
                name_filters=[nf for nf in self._api.get_save_file_filters()],
                selected_name_filter=self._api.get_save_file_filter(),
            )

            if file_path:
                try:
                    self._api.save_product(current.row(), file_path, file_type=name_filter)
                except Exception as err:
                    logger.exception(err)
                    ExceptionDialog.show_exception('File Writer', err)
        else:
            logger.error('No current item!')

    def _sync_current_product_to_settings(self) -> None:
        item_index = self._get_current_item_index()

        if item_index < 0:
            logger.warning('No current item!')
        else:
            item = self._repository[item_index]
            item.sync_to_settings()

    def _duplicate_current_product(self) -> None:
        current = self._table_proxy_model.mapToSource(self._view.table_view.currentIndex())

        if current.isValid():
            like_item = self._repository[current.row()]
            self._api.insert_product(like_item.get_product())
        else:
            logger.error('No current item!')

    def _edit_current_product(self) -> None:
        current = self._table_proxy_model.mapToSource(self._view.table_view.currentIndex())

        if current.isValid():
            product = self._repository[current.row()]
            ProductEditorViewController.edit_product(self._diffraction_api, product, self._view)
        else:
            logger.error('No current item!')

    def _remove_current_product(self) -> None:
        current = self._table_proxy_model.mapToSource(self._view.table_view.currentIndex())

        if current.isValid():
            self._repository.remove_product(current.row())
        else:
            logger.error('No current item!')

    def _update_enabled_buttons(self, current: QModelIndex, previous: QModelIndex) -> None:
        enabled = current.isValid()
        self._duplicate_action.setEnabled(enabled)
        self._view.button_box.save_button.setEnabled(enabled)
        self._view.button_box.edit_button.setEnabled(enabled)
        self._view.button_box.remove_button.setEnabled(enabled)

    def _update_info_text(self) -> None:
        info_text = self._repository.get_info_text()
        self._view.info_label.setText(info_text)

    def handle_item_inserted(self, index: int, item: ProductRepositoryItem) -> None:
        parent = QModelIndex()
        self._table_model.beginInsertRows(parent, index, index)
        self._table_model.endInsertRows()
        self._update_info_text()

    def handle_metadata_changed(self, index: int, item: MetadataRepositoryItem) -> None:
        top_left = self._table_model.index(index, 0)
        bottom_right = self._table_model.index(index, self._table_model.columnCount() - 1)
        self._table_model.dataChanged.emit(top_left, bottom_right, [Qt.ItemDataRole.DisplayRole])
        self._update_info_text()

    def handle_probe_positions_changed(
        self, index: int, item: ProbePositionsRepositoryItem
    ) -> None:
        self._update_info_text()

    def handle_probe_changed(self, index: int, item: ProbeRepositoryItem) -> None:
        self._update_info_text()

    def handle_object_changed(self, index: int, item: ObjectRepositoryItem) -> None:
        self._update_info_text()

    def handle_losses_changed(self, index: int, losses: Sequence[LossValue]) -> None:
        self._update_info_text()

    def handle_item_removed(self, index: int, item: ProductRepositoryItem) -> None:
        parent = QModelIndex()
        self._table_model.beginRemoveRows(parent, index, index)
        self._table_model.endRemoveRows()
        self._update_info_text()



from pathlib import Path
import csv

from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePositionFileWriter,
    ProbePosition,
    ProbePositionParseError,
)


class DelimitedPositionFileReader(ProbePositionFileReader):
    def __init__(self, delimiter: str, *, swap_xy: bool) -> None:
        self._delimiter = delimiter
        self._swap_xy = swap_xy

    def read(self, file_path: Path) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()

        if self._swap_xy:
            xcol = 1
            ycol = 0
        else:
            xcol = 0
            ycol = 1

        with file_path.open(newline='') as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=self._delimiter)

            for idx, row in enumerate(csv_reader):
                if row[0].startswith('#'):
                    continue

                if len(row) < 2:
                    raise ProbePositionParseError('Bad number of columns!')

                point = ProbePosition(idx, float(row[xcol]), float(row[ycol]))
                point_list.append(point)

        return ProbePositionSequence(point_list)


class DelimitedPositionFileWriter(ProbePositionFileWriter):
    def __init__(self, delimiter: str, swap_xy: bool) -> None:
        self._delimiter = delimiter
        self._swap_xy = swap_xy

    def write(self, file_path: Path, positions: ProbePositionSequence) -> None:
        with file_path.open(mode='wt') as csv_file:
            for point in positions:
                x = point.coordinate_x_m
                y = point.coordinate_y_m
                line = (
                    f'{y}{self._delimiter}{x}\n' if self._swap_xy else f'{x}{self._delimiter}{y}\n'
                )
                csv_file.write(line)


def register_plugins(registry: PluginRegistry) -> None:
    registry.probe_position_file_readers.register_plugin(
        DelimitedPositionFileReader(' ', swap_xy=False),
        simple_name='TXT',
        display_name='Space-Separated Values Files (*.txt)',
    )
    registry.probe_position_file_writers.register_plugin(
        DelimitedPositionFileWriter(' ', swap_xy=False),
        simple_name='TXT',
        display_name='Space-Separated Values Files (*.txt)',
    )
    registry.probe_position_file_readers.register_plugin(
        DelimitedPositionFileReader(',', swap_xy=True),
        simple_name='CSV',
        display_name='Comma-Separated Values Files (*.csv)',
    )
    registry.probe_position_file_writers.register_plugin(
        DelimitedPositionFileWriter(',', swap_xy=True),
        simple_name='CSV',
        display_name='Comma-Separated Values Files (*.csv)',
    )



from .affine import PtyChiAffineDegreesOfFreedomBitField
from .core import PtyChiReconstructorLibrary
from .device import PtyChiDeviceRepository
from .enums import PtyChiEnumerators
from .settings import (
    PtyChiAutodiffSettings,
    PtyChiDMSettings,
    PtyChiLSQMLSettings,
    PtyChiOPRSettings,
    PtyChiObjectSettings,
    PtyChiPIESettings,
    PtyChiProbePositionSettings,
    PtyChiProbeSettings,
    PtyChiSettings,
)

__all__ = [
    'PtyChiAffineDegreesOfFreedomBitField',
    'PtyChiAutodiffSettings',
    'PtyChiDMSettings',
    'PtyChiDeviceRepository',
    'PtyChiEnumerators',
    'PtyChiLSQMLSettings',
    'PtyChiOPRSettings',
    'PtyChiObjectSettings',
    'PtyChiPIESettings',
    'PtyChiProbePositionSettings',
    'PtyChiProbeSettings',
    'PtyChiReconstructorLibrary',
    'PtyChiSettings',
]



from collections.abc import Sequence
from typing import overload
import logging
import sys

from ptychodus.api.common import BYTES_PER_MEGABYTE

from .item import ProductRepositoryItem, ProductRepositoryItemObserver, ProductRepositoryObserver

logger = logging.getLogger(__name__)


class ProductRepository(Sequence[ProductRepositoryItem], ProductRepositoryItemObserver):
    def __init__(self) -> None:
        super().__init__()
        self._item_list: list[ProductRepositoryItem] = []
        self._observer_list: list[ProductRepositoryObserver] = []

    @overload
    def __getitem__(self, index: int) -> ProductRepositoryItem: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[ProductRepositoryItem]: ...

    def __getitem__(
        self, index: int | slice
    ) -> ProductRepositoryItem | Sequence[ProductRepositoryItem]:
        return self._item_list[index]

    def __len__(self) -> int:
        return len(self._item_list)

    def create_unique_name(self, candidate_name: str) -> str:
        reserved_names = set([item.get_name() for item in self._item_list])
        name = candidate_name or 'Unnamed'
        match = 0

        while name in reserved_names:
            match += 1
            name = f'{candidate_name}-{match}'

        return name

    def _update_indexes(self) -> None:
        for index, item in enumerate(self._item_list):
            item._index = index

    def insert_product(self, item: ProductRepositoryItem) -> int:
        index = len(self._item_list)
        self._item_list.append(item)

        self._update_indexes()

        for observer in self._observer_list:
            observer.handle_item_inserted(index, item)

        return index

    def remove_product(self, index: int) -> None:
        try:
            item = self._item_list.pop(index)
        except IndexError:
            logger.debug(f'Failed to remove product item {index}!')
            return

        self._update_indexes()

        for observer in self._observer_list:
            observer.handle_item_removed(index, item)

    def get_info_text(self) -> str:
        size_MB = sum(sys.getsizeof(prod) for prod in self._item_list) / BYTES_PER_MEGABYTE  # noqa: N806
        return f'Total: {len(self)} [{size_MB:.2f}MB]'

    def add_observer(self, observer: ProductRepositoryObserver) -> None:
        if observer not in self._observer_list:
            self._observer_list.append(observer)

    def remove_observer(self, observer: ProductRepositoryObserver) -> None:
        try:
            self._observer_list.remove(observer)
        except ValueError:
            pass

    def handle_metadata_changed(self, item: ProductRepositoryItem) -> None:
        metadata = item.get_metadata_item()
        index = item._index

        if index < 0:
            logger.warning(f'Failed to look up index for "{item.get_name()}"!')
            return

        for observer in self._observer_list:
            observer.handle_metadata_changed(index, metadata)

    def handle_probe_positions_changed(self, item: ProductRepositoryItem) -> None:
        index = item._index
        scan = item.get_probe_positions_item()

        if index < 0:
            logger.warning(f'Failed to look up index for "{item.get_name()}"!')
            return

        for observer in self._observer_list:
            observer.handle_probe_positions_changed(index, scan)

    def handle_probe_changed(self, item: ProductRepositoryItem) -> None:
        index = item._index
        probe = item.get_probe_item()

        if index < 0:
            logger.warning(f'Failed to look up index for "{item.get_name()}"!')
            return

        for observer in self._observer_list:
            observer.handle_probe_changed(index, probe)

    def handle_object_changed(self, item: ProductRepositoryItem) -> None:
        index = item._index
        object_ = item.get_object_item()

        if index < 0:
            logger.warning(f'Failed to look up index for "{item.get_name()}"!')
            return

        for observer in self._observer_list:
            observer.handle_object_changed(index, object_)

    def handle_losses_changed(self, item: ProductRepositoryItem) -> None:
        index = item._index
        losses = item.get_losses()

        if index < 0:
            logger.warning(f'Failed to look up index for "{item.get_name()}"!')
            return

        for observer in self._observer_list:
            observer.handle_losses_changed(index, losses)



from __future__ import annotations

from ptychodus.api.observer import Observable
from ptychodus.api.diffraction import DiffractionMetadata

from .diffraction import (
    DetectorSettings,
    DiffractionDatasetObserver,
    AssembledDiffractionDataset,
    DiffractionSettings,
)
from .product import ProductSettings


class MetadataPresenter(Observable, DiffractionDatasetObserver):
    def __init__(
        self,
        detector_settings: DetectorSettings,
        diffraction_settings: DiffractionSettings,
        dataset: AssembledDiffractionDataset,
        product_settings: ProductSettings,
    ) -> None:
        super().__init__()
        self._detector_settings = detector_settings
        self._diffraction_settings = diffraction_settings
        self._dataset = dataset
        self._product_settings = product_settings

        dataset.add_observer(self)

    @property
    def _metadata(self) -> DiffractionMetadata:
        return self._dataset.get_metadata()

    def can_sync_detector_extent(self) -> bool:
        return self._metadata.detector_extent is not None

    def sync_detector_extent(self) -> None:
        detector_extent = self._metadata.detector_extent

        if detector_extent:
            self._detector_settings.width_px.set_value(detector_extent.width_px)
            self._detector_settings.height_px.set_value(detector_extent.height_px)

    def can_sync_detector_pixel_size(self) -> bool:
        return self._metadata.detector_pixel_geometry is not None

    def sync_detector_pixel_size(self) -> None:
        pixel_geometry = self._metadata.detector_pixel_geometry

        if pixel_geometry:
            self._detector_settings.pixel_width_m.set_value(pixel_geometry.width_m)
            self._detector_settings.pixel_height_m.set_value(pixel_geometry.height_m)

    def can_sync_detector_bit_depth(self) -> bool:
        return self._metadata.detector_bit_depth is not None

    def sync_detector_bit_depth(self) -> None:
        bit_depth = self._metadata.detector_bit_depth

        if bit_depth:
            self._detector_settings.bit_depth.set_value(bit_depth)

    def can_sync_pattern_crop_center(self) -> bool:
        return self._metadata.crop_center is not None or self._metadata.detector_extent is not None

    def can_sync_pattern_crop_extent(self) -> bool:
        return self._metadata.detector_extent is not None

    def sync_pattern_crop(self, sync_center: bool, sync_extent: bool) -> None:
        if sync_center:
            crop_center = self._metadata.crop_center

            if crop_center:
                self._diffraction_settings.crop_center_x_px.set_value(crop_center.position_x_px)
                self._diffraction_settings.crop_center_y_px.set_value(crop_center.position_y_px)
            elif self._metadata.detector_extent:
                self._diffraction_settings.crop_center_x_px.set_value(
                    int(self._metadata.detector_extent.width_px) // 2
                )
                self._diffraction_settings.crop_center_y_px.set_value(
                    int(self._metadata.detector_extent.height_px) // 2
                )

        if sync_extent and self._metadata.detector_extent:
            center_x = self._diffraction_settings.crop_center_x_px.get_value()
            center_y = self._diffraction_settings.crop_center_y_px.get_value()

            extent_x = int(self._metadata.detector_extent.width_px)
            extent_y = int(self._metadata.detector_extent.height_px)

            max_radius_x = min(center_x, extent_x - center_x)
            max_radius_y = min(center_y, extent_y - center_y)
            max_radius = min(max_radius_x, max_radius_y)
            crop_diameter = 1

            while crop_diameter < max_radius:
                crop_diameter <<= 1

            self._diffraction_settings.crop_width_px.set_value(crop_diameter)
            self._diffraction_settings.crop_height_px.set_value(crop_diameter)

    def can_sync_probe_energy(self) -> bool:
        return self._metadata.probe_energy_eV is not None

    def sync_probe_energy(self) -> None:
        energy_eV = self._metadata.probe_energy_eV  # noqa: N806

        if energy_eV:
            self._product_settings.probe_energy_eV.set_value(energy_eV)

    def can_sync_probe_photon_count(self) -> bool:
        return self._metadata.probe_photon_count is not None

    def sync_probe_photon_count(self) -> None:
        photon_count = self._metadata.probe_photon_count

        if photon_count:
            self._product_settings.probe_photon_count.set_value(photon_count)

    def can_sync_exposure_time(self) -> bool:
        return self._metadata.exposure_time_s is not None

    def sync_exposure_time(self) -> None:
        exposure_time_s = self._metadata.exposure_time_s

        if exposure_time_s:
            self._product_settings.exposure_time_s.set_value(exposure_time_s)

    def can_sync_detector_distance(self) -> bool:
        return self._metadata.detector_distance_m is not None

    def sync_detector_distance(self) -> None:
        distance_m = self._metadata.detector_distance_m

        if distance_m:
            self._product_settings.detector_distance_m.set_value(distance_m)

    def handle_array_inserted(self, index: int) -> None:
        pass

    def handle_array_changed(self, index: int) -> None:
        pass

    def handle_dataset_reloaded(self) -> None:
        self.notify_observers()



from __future__ import annotations
from collections.abc import Iterator
import logging

from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.observer import Observable, Observer
from ptychodus.api.plugins import PluginChooser
from ptychodus.api.visualization import (
    NumberArrayType,
    RealArrayType,
    VisualizationProduct,
)

from .color_axis import ColorAxis
from .color_model_renderer import CylindricalColorModelRenderer
from .colormap import ColormapParameter
from .colormap_renderer import ColormapRenderer
from .components import (
    AmplitudeArrayComponent,
    ImaginaryArrayComponent,
    PhaseInRadiansArrayComponent,
    RealArrayComponent,
    UnwrappedPhaseInRadiansArrayComponent,
)

from .renderer import Renderer
from .transformation import ScalarTransformationParameter

logger = logging.getLogger(__name__)


class VisualizationEngine(Observable, Observer):
    def __init__(self, *, is_complex: bool) -> None:
        super().__init__()
        self._renderer_chooser = PluginChooser[Renderer]()
        self._transformation = ScalarTransformationParameter()
        self._color_axis = ColorAxis()
        acyclic_colormap = ColormapParameter(is_cyclic=False)

        self._renderer_chooser.register_plugin(
            ColormapRenderer(
                RealArrayComponent(),
                self._transformation,
                self._color_axis,
                acyclic_colormap,
            ),
            display_name='Real',
        )

        if is_complex:
            amplitude_component = AmplitudeArrayComponent()
            phase_component = PhaseInRadiansArrayComponent()
            cyclic_colormap = ColormapParameter(is_cyclic=True)

            self._renderer_chooser.register_plugin(
                ColormapRenderer(
                    ImaginaryArrayComponent(),
                    self._transformation,
                    self._color_axis,
                    acyclic_colormap,
                ),
                display_name='Imaginary',
            )
            self._renderer_chooser.register_plugin(
                ColormapRenderer(
                    amplitude_component,
                    self._transformation,
                    self._color_axis,
                    acyclic_colormap,
                ),
                display_name='Amplitude',
            )
            self._renderer_chooser.register_plugin(
                ColormapRenderer(
                    phase_component,
                    self._transformation,
                    self._color_axis,
                    cyclic_colormap,
                ),
                display_name='Phase',
            )
            self._renderer_chooser.register_plugin(
                ColormapRenderer(
                    UnwrappedPhaseInRadiansArrayComponent(),
                    self._transformation,
                    self._color_axis,
                    acyclic_colormap,
                ),
                display_name='Phase (Unwrapped)',
            )
            self._renderer_chooser.register_plugin(
                CylindricalColorModelRenderer(
                    amplitude_component,
                    phase_component,
                    self._transformation,
                    self._color_axis,
                ),
                display_name='Complex',
            )
            self._renderer_chooser.set_current_plugin('Complex')

        self._renderer_chooser.add_observer(self)
        self._renderer_plugin = self._renderer_chooser.get_current_plugin()
        self._renderer_plugin.strategy.add_observer(self)

    def renderers(self) -> Iterator[str]:
        for plugin in self._renderer_chooser:
            yield plugin.display_name

    def get_renderer(self) -> str:
        return self._renderer_plugin.display_name

    def set_renderer(self, value: str) -> None:
        self._renderer_chooser.set_current_plugin(value)

    def is_renderer_cyclic(self) -> bool:
        return self._renderer_plugin.strategy.is_cyclic()

    def transformations(self) -> Iterator[str]:
        return self._transformation.choices()

    def get_transformation(self) -> str:
        return self._transformation.get_value()

    def set_transformation(self, value: str) -> None:
        self._transformation.set_value(value)

    def variants(self) -> Iterator[str]:
        return self._renderer_plugin.strategy.variants()

    def get_variant(self) -> str:
        return self._renderer_plugin.strategy.get_variant()

    def set_variant(self, value: str) -> None:
        return self._renderer_plugin.strategy.set_variant(value)

    def get_min_display_value(self) -> float:
        return self._color_axis.lower.get_value()

    def set_min_display_value(self, value: float) -> None:
        self._color_axis.lower.set_value(value)

    def get_max_display_value(self) -> float:
        return self._color_axis.upper.get_value()

    def set_max_display_value(self, value: float) -> None:
        self._color_axis.upper.set_value(value)

    def set_display_value_range(self, lower: float, upper: float) -> None:
        self._color_axis.set_range(lower, upper)

    def colorize(self, array: NumberArrayType) -> RealArrayType:
        return self._renderer_plugin.strategy.colorize(array)

    def render(
        self,
        array: NumberArrayType,
        pixel_geometry: PixelGeometry,
        *,
        autoscale_color_axis: bool,
    ) -> VisualizationProduct:
        return self._renderer_plugin.strategy.render(
            array, pixel_geometry, autoscale_color_axis=autoscale_color_axis
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._renderer_chooser:
            self._renderer_plugin.strategy.remove_observer(self)
            self._renderer_plugin = self._renderer_chooser.get_current_plugin()
            self._renderer_plugin.strategy.add_observer(self)
            self.notify_observers()
        elif observable is self._renderer_plugin.strategy:
            self.notify_observers()



from ptychodus.api.geometry import ImageExtent, Interval, PixelGeometry
from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import BooleanParameter, IntegerParameter, RealParameter
from ptychodus.api.diffraction import CropCenter

from .processor import (
    DiffractionPatternBinning,
    DiffractionPatternCrop,
    DiffractionPatternFilterValues,
    DiffractionPatternPadding,
    DiffractionPatternProcessor,
)
from .settings import DetectorSettings, DiffractionSettings


class PatternAxisSizer(Observable, Observer):
    def __init__(
        self,
        detector_size: IntegerParameter,
        detector_pixel_size_m: RealParameter,
        crop_enabled: BooleanParameter,
        crop_size: IntegerParameter,
        crop_center: IntegerParameter,
        binning_enabled: BooleanParameter,
        bin_size: IntegerParameter,
        padding_enabled: BooleanParameter,
        pad_size: IntegerParameter,
    ) -> None:
        super().__init__()
        self._detector_size = detector_size
        self._detector_pixel_size_m = detector_pixel_size_m
        self._crop_enabled = crop_enabled
        self._crop_size = crop_size
        self._crop_center = crop_center
        self._binning_enabled = binning_enabled
        self._bin_size = bin_size
        self._padding_enabled = padding_enabled
        self._pad_size = pad_size

        detector_size.add_observer(self)
        detector_pixel_size_m.add_observer(self)
        crop_enabled.add_observer(self)
        crop_size.add_observer(self)
        crop_center.add_observer(self)
        binning_enabled.add_observer(self)
        bin_size.add_observer(self)
        padding_enabled.add_observer(self)
        pad_size.add_observer(self)

    def get_detector_size(self) -> int:
        return self._detector_size.get_value()

    def get_crop_size_limits(self) -> Interval[int]:
        return Interval[int](1, self.get_detector_size())

    def get_crop_size(self) -> int:
        if self._crop_enabled.get_value():
            limits = self.get_crop_size_limits()
            return limits.clamp(self._crop_size.get_value())

        return self.get_detector_size()

    def _get_safe_crop_center(self) -> int:
        xmin = (self.get_crop_size() + 1) // 2
        xmax = self.get_detector_size() - 1 - xmin
        limits = Interval[int](xmin, xmax)
        return limits.clamp(self._crop_center.get_value())

    def get_crop_center_limits(self) -> Interval[int]:
        return Interval[int](1, self.get_detector_size())

    def get_crop_center(self) -> int:
        limits = self.get_crop_center_limits()
        return limits.clamp(self._crop_center.get_value())

    def get_bin_size_limits(self) -> Interval[int]:
        return Interval[int](1, self.get_crop_size())

    def get_bin_size(self) -> int:
        if self._binning_enabled.get_value():
            limits = self.get_bin_size_limits()
            return limits.clamp(self._bin_size.get_value())

        return 1

    def validate_bin_size(self) -> None:
        crop_size = self.get_crop_size()
        bin_size = self.get_bin_size()

        if crop_size % bin_size != 0:
            raise ValueError(f'Invalid binning size! ({crop_size=}, {bin_size=})')

    def get_pad_size(self) -> int:
        if self._padding_enabled.get_value():
            return self._pad_size.get_value()

        return 0

    def get_processed_size(self) -> int:
        return self.get_crop_size() // self.get_bin_size() + self.get_pad_size()

    def get_processed_pixel_size_m(self) -> float:
        return self.get_bin_size() * self._detector_pixel_size_m.get_value()

    def get_processed_size_m(self) -> float:
        return self.get_processed_size() * self.get_processed_pixel_size_m()

    def _update(self, observable: Observable) -> None:
        if observable in (
            self._detector_size,
            self._detector_pixel_size_m,
            self._crop_enabled,
            self._crop_size,
            self._crop_center,
            self._binning_enabled,
            self._bin_size,
            self._padding_enabled,
            self._pad_size,
        ):
            self.notify_observers()


class PatternSizer(Observable, Observer):
    def __init__(
        self, detector_settings: DetectorSettings, diffraction_settings: DiffractionSettings
    ) -> None:
        super().__init__()
        self._diffraction_settings = diffraction_settings
        self.axis_x = PatternAxisSizer(
            detector_settings.width_px,
            detector_settings.pixel_width_m,
            diffraction_settings.crop_enabled,
            diffraction_settings.crop_width_px,
            diffraction_settings.crop_center_x_px,
            diffraction_settings.binning_enabled,
            diffraction_settings.bin_size_x,
            diffraction_settings.padding_enabled,
            diffraction_settings.pad_x,
        )
        self.axis_y = PatternAxisSizer(
            detector_settings.height_px,
            detector_settings.pixel_height_m,
            diffraction_settings.crop_enabled,
            diffraction_settings.crop_height_px,
            diffraction_settings.crop_center_y_px,
            diffraction_settings.binning_enabled,
            diffraction_settings.bin_size_y,
            diffraction_settings.padding_enabled,
            diffraction_settings.pad_y,
        )

        self.axis_x.add_observer(self)
        self.axis_y.add_observer(self)

    def get_detector_extent(self) -> ImageExtent:
        return ImageExtent(
            width_px=self.axis_x.get_detector_size(),
            height_px=self.axis_y.get_detector_size(),
        )

    def get_processed_width_m(self) -> float:
        return self.axis_x.get_processed_size_m()

    def get_processed_height_m(self) -> float:
        return self.axis_y.get_processed_size_m()

    def get_processed_image_extent(self) -> ImageExtent:
        return ImageExtent(
            width_px=self.axis_x.get_processed_size(),
            height_px=self.axis_y.get_processed_size(),
        )

    def get_processed_pixel_geometry(self) -> PixelGeometry:
        return PixelGeometry(
            width_m=self.axis_x.get_processed_pixel_size_m(),
            height_m=self.axis_y.get_processed_pixel_size_m(),
        )

    def get_processor(self) -> DiffractionPatternProcessor:
        value_lower_bound: int | None = None
        value_upper_bound: int | None = None
        crop: DiffractionPatternCrop | None = None
        binning: DiffractionPatternBinning | None = None
        padding: DiffractionPatternPadding | None = None

        if self._diffraction_settings.value_upper_bound_enabled.get_value():
            value_lower_bound = self._diffraction_settings.value_lower_bound.get_value()

        if self._diffraction_settings.value_upper_bound_enabled.get_value():
            value_upper_bound = self._diffraction_settings.value_upper_bound.get_value()

        filter_values = DiffractionPatternFilterValues(
            lower_bound=value_lower_bound,
            upper_bound=value_upper_bound,
        )

        if self._diffraction_settings.crop_enabled.get_value():
            crop = DiffractionPatternCrop(
                center=CropCenter(
                    self.axis_x._get_safe_crop_center(),
                    self.axis_y._get_safe_crop_center(),
                ),
                extent=ImageExtent(
                    self.axis_x.get_crop_size(),
                    self.axis_y.get_crop_size(),
                ),
            )

        if self._diffraction_settings.binning_enabled.get_value():
            self.axis_x.validate_bin_size()
            self.axis_y.validate_bin_size()
            binning = DiffractionPatternBinning(
                bin_size_x=self.axis_x.get_bin_size(),
                bin_size_y=self.axis_y.get_bin_size(),
            )

        if self._diffraction_settings.padding_enabled.get_value():
            padding = DiffractionPatternPadding(
                pad_x=self.axis_x.get_pad_size(),
                pad_y=self.axis_y.get_pad_size(),
            )

        return DiffractionPatternProcessor(
            filter_values=filter_values,
            crop=crop,
            binning=binning,
            padding=padding,
            hflip=self._diffraction_settings.hflip.get_value(),
            vflip=self._diffraction_settings.vflip.get_value(),
            transpose=self._diffraction_settings.transpose.get_value(),
        )

    def _update(self, observable: Observable) -> None:
        if observable in (self.axis_x, self.axis_y):
            self.notify_observers()



from pathlib import Path

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class ProbeSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('Probe')
        self._group.add_observer(self)

        self.builder = self._group.create_string_parameter('Builder', 'Disk')
        self.file_path = self._group.create_path_parameter('FilePath', Path('/path/to/probe.npy'))
        self.file_type = self._group.create_string_parameter('FileType', 'NPY')

        self.num_incoherent_modes = self._group.create_integer_parameter(
            'NumberOfIncoherentModes', 1, minimum=1
        )
        self.orthogonalize_incoherent_modes = self._group.create_boolean_parameter(
            'OrthogonalizeIncoherentModes', True
        )
        self.incoherent_mode_decay_type = self._group.create_string_parameter(
            'IncoherentModeDecayType', 'Polynomial'
        )
        self.incoherent_mode_decay_ratio = self._group.create_real_parameter(
            'IncoherentModeDecayRatio', 1.0, minimum=0.0, maximum=1.0
        )
        self.num_coherent_modes = self._group.create_integer_parameter(
            'NumberOfCoherentModes', 1, minimum=1
        )

        self.disk_diameter_m = self._group.create_real_parameter(
            'DiskDiameterInMeters', 1e-6, minimum=0.0
        )
        self.rectangle_width_m = self._group.create_real_parameter(
            'RectangleWidthInMeters', 1e-6, minimum=0.0
        )
        self.rectangle_height_m = self._group.create_real_parameter(
            'RectangleHeightInMeters', 1e-6, minimum=0.0
        )

        self.super_gaussian_annular_radius_m = self._group.create_real_parameter(
            'SuperGaussianAnnularRadiusInMeters', 0, minimum=0.0
        )
        self.super_gaussian_width_m = self._group.create_real_parameter(
            'SuperGaussianWidthInMeters', 400e-6, minimum=0.0
        )
        self.super_gaussian_order_parameter = self._group.create_real_parameter(
            'SuperGaussianOrderParameter', 1, minimum=1.0
        )

        self.zone_plate_diameter_m = self._group.create_real_parameter(
            'ZonePlateDiameterInMeters', 180e-6, minimum=0.0
        )
        self.outermost_zone_width_m = self._group.create_real_parameter(
            'OutermostZoneWidthInMeters', 50e-9, minimum=0.0
        )
        self.central_beamstop_diameter_m = self._group.create_real_parameter(
            'CentralBeamstopDiameterInMeters', 60e-6, minimum=0.0
        )
        self.defocus_distance_m = self._group.create_real_parameter('DefocusDistanceInMeters', 0.0)

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class ProbePropagationSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('ProbePropagation')
        self._group.add_observer(self)

        self.begin_coordinate_m = self._group.create_real_parameter(
            'BeginCoordinateInMeters', -1e-3
        )
        self.end_coordinate_m = self._group.create_real_parameter('EndCoordinateInMeters', 1e-3)
        self.num_steps = self._group.create_integer_parameter('NumberOfSteps', 100)

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class AffineTransformEstimatorSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('AffineTransformEstimator')
        self._group.add_observer(self)

        self.num_shuffles = self._group.create_integer_parameter('NumberOfShuffles', 10, minimum=1)
        self.inlier_threshold = self._group.create_real_parameter(
            'InlierThreshold', 1e-6, minimum=0.0
        )
        self.min_inliers = self._group.create_integer_parameter('MinimumInliers', 10, minimum=3)

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from PyQt5.QtWidgets import QHBoxLayout, QLabel, QSpinBox, QWidget

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import IntegerParameter, StringParameter

from ...model.ptychi import PtyChiEnumerators
from ..parametric import (
    ComboBoxParameterViewController,
    SpinBoxParameterViewController,
    ParameterViewController,
)

__all__ = [
    'PtyChiOptimizationPlanViewController',
    'PtyChiOptimizerParameterViewController',
]


class PtyChiStopSpinBoxParameterViewController(ParameterViewController, Observer):
    def __init__(
        self, stop: IntegerParameter, num_epochs: IntegerParameter, *, tool_tip: str = ''
    ) -> None:
        super().__init__()
        self._stop = stop
        self._num_epochs = num_epochs
        self._widget = QSpinBox()

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self._sync_model_to_view()
        self._widget.valueChanged.connect(self._sync_view_to_model)
        stop.add_observer(self)
        num_epochs.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def _sync_view_to_model(self, value: int) -> None:
        num_epochs = self._num_epochs.get_value()
        self._stop.set_value(value if value < num_epochs else -1)

    def _sync_model_to_view(self) -> None:
        num_epochs = self._num_epochs.get_value()
        stop = self._stop.get_value()

        self._widget.blockSignals(True)
        self._widget.setRange(0, num_epochs)
        self._widget.setValue(num_epochs if stop < 0 else stop)
        self._widget.blockSignals(False)

    def _update(self, observable: Observable) -> None:
        if observable in (self._stop, self._num_epochs):
            self._sync_model_to_view()


class PtyChiOptimizationPlanViewController(ParameterViewController):
    def __init__(
        self,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        num_epochs: IntegerParameter,
    ) -> None:
        super().__init__()
        self._start_view_controller = SpinBoxParameterViewController(
            start, tool_tip='Iteration to start optimizing'
        )
        self._stop_view_controller = PtyChiStopSpinBoxParameterViewController(
            stop, num_epochs, tool_tip='Iteration to stop optimizing'
        )
        self._stride_view_controller = SpinBoxParameterViewController(
            stride, tool_tip='Number of iterations between updates'
        )
        self._widget = QWidget()

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(QLabel('Start'), 0)
        layout.addWidget(self._start_view_controller.get_widget(), 1)
        layout.addWidget(QLabel('Stop'), 0)
        layout.addWidget(self._stop_view_controller.get_widget(), 1)
        layout.addWidget(QLabel('Stride'), 0)
        layout.addWidget(self._stride_view_controller.get_widget(), 1)
        self._widget.setLayout(layout)

    def get_widget(self) -> QWidget:
        return self._widget


class PtyChiOptimizerParameterViewController(ComboBoxParameterViewController):
    def __init__(self, parameter: StringParameter, enumerators: PtyChiEnumerators) -> None:
        super().__init__(parameter, enumerators.optimizers(), tool_tip='Name of the optimizer.')
